import asyncio
from typing import List, Dict, Any, Optional, Union
from concurrent.futures import ThreadPoolExecutor

from grammared_language.clients.base_client import BaseClient
from grammared_language.language_tool.output_models import LanguageToolRemoteResult
from grammared_language.utils.config_parser import create_clients_from_config


class AsyncMultiClient:
    """Client that runs multiple grammar correction clients asynchronously."""
    
    def __init__(self, clients: Optional[List[BaseClient]] = None, config_path: Optional[str] = None, max_workers: Optional[int] = None):
        """
        Initialize AsyncMultiClient.
        
        Args:
            clients: List of pre-initialized client instances
            config_path: Path to model_config.yaml to initialize clients from
            max_workers: Maximum number of worker threads for executor (default: None = auto)
        """
        self.clients: List[BaseClient] = clients or []
        self._executor = ThreadPoolExecutor(max_workers=max_workers)
        
        if config_path and not clients:
            self.clients = create_clients_from_config(config_path)
    
    def add_client(self, client: BaseClient):
        """Add a client to the list."""
        self.clients.append(client)
    
    async def _predict_async(self, text: str) -> List[LanguageToolRemoteResult]:
        """
        Run all clients asynchronously on the input text (internal async method).
        
        Args:
            text: Input text to process
            
        Returns:
            List of LanguageToolRemoteResult from each client
        """
        loop = asyncio.get_running_loop()
        tasks = [loop.run_in_executor(self._executor, client.predict, text) for client in self.clients]
        results = await asyncio.gather(*tasks, return_exceptions=True)
        
        # Filter out exceptions and return valid results
        valid_results = []
        for i, result in enumerate(results):
            if isinstance(result, Exception):
                print(f"Client {i} failed with error: {result}")
            else:
                valid_results.append(result)
        
        return valid_results
    
    async def _predict_batch_async(self, texts: List[str]) -> List[List[LanguageToolRemoteResult]]:
        """
        Process multiple texts concurrently across all clients (internal async method).
        Leverages batch processing capability of each client for efficiency.
        
        Args:
            texts: List of input texts to process
            
        Returns:
            List of results for each text, where each result is a list from all clients
        """
        loop = asyncio.get_running_loop()
        # Run each client on the entire batch concurrently
        tasks = [loop.run_in_executor(self._executor, client.predict, texts) for client in self.clients]
        client_results = await asyncio.gather(*tasks, return_exceptions=True)
        
        # Reorganize results: from [client_results] to [text_results]
        # client_results[i] = list of results from client i for all texts
        # We want: results[j] = list of results from all clients for text j
        num_texts = len(texts)
        text_results = [[] for _ in range(num_texts)]
        
        for client_idx, result in enumerate(client_results):
            if isinstance(result, Exception):
                print(f"Client {client_idx} failed with error: {result}")
                # Add empty results for all texts from this failed client
                for text_idx in range(num_texts):
                    text_results[text_idx].append(LanguageToolRemoteResult(
                        language="English",
                        languageCode="en-US",
                        matches=[]
                    ))
            else:
                # result is a list of LanguageToolRemoteResult for this client
                for text_idx, text_result in enumerate(result):
                    text_results[text_idx].append(text_result)
        
        return text_results
    
    async def _predict_with_merge_async(self, text: str) -> LanguageToolRemoteResult:
        """
        Run all clients asynchronously and merge results (internal async method).
        
        Args:
            text: Input text to process
            
        Returns:
            Merged LanguageToolRemoteResult
        """
        results = await self._predict_async(text)
        
        if not results:
            return LanguageToolRemoteResult(
                language="English",
                languageCode="en-US",
                matches=[]
            )
        
        # Merge matches from all results
        merged_matches = []
        seen_replacements = set()
        
        for result in results:
            for match in result.matches:
                # Create unique key for deduplication based on offset and length
                # Use suggested_replacements to get replacement strings
                replacements_list = []
                if match.suggested_replacements:
                    replacements_list = [r.replacement for r in match.suggested_replacements]
                elif match.suggestions:
                    replacements_list = match.suggestions
                    
                key = (match.offset, match.length, tuple(replacements_list))
                if key not in seen_replacements:
                    seen_replacements.add(key)
                    merged_matches.append(match)
        
        # Sort by offset
        merged_matches.sort(key=lambda m: m.offset)
        
        return LanguageToolRemoteResult(
            language="English",
            languageCode="en-US",
            matches=merged_matches
        )
    
    async def _predict_batch_with_merge_async(self, texts: List[str]) -> List[LanguageToolRemoteResult]:
        """
        Process multiple texts and merge results for each text (internal async method).
        
        Args:
            texts: List of input texts to process
            
        Returns:
            List of merged LanguageToolRemoteResult for each text
        """
        tasks = [self._predict_with_merge_async(text) for text in texts]
        results = await asyncio.gather(*tasks, return_exceptions=True)
        
        # Filter out exceptions and return valid results
        valid_results = []
        for i, result in enumerate(results):
            if isinstance(result, Exception):
                print(f"Text {i} failed with error: {result}")
                # Return empty result for failed text
                valid_results.append(LanguageToolRemoteResult(
                    language="English",
                    languageCode="en-US",
                    matches=[]
                ))
            else:
                valid_results.append(result)
        
        return valid_results
    
    def predict(self, text: str) -> List[LanguageToolRemoteResult]:
        """
        Run all clients and return list of results from each client.
        
        Args:
            text: Input text to process
            
        Returns:
            List of LanguageToolRemoteResult from each client
        """
        try:
            loop = asyncio.get_running_loop()
            # If loop is already running, we can't use run_until_complete or asyncio.run in this thread
            # Create a temporary executor to run the async task in a new thread
            with ThreadPoolExecutor(max_workers=1) as executor:
                future = executor.submit(asyncio.run, self._predict_async(text))
                return future.result()
        except RuntimeError:
            # No event loop exists or is running in this thread, create one
            return asyncio.run(self._predict_async(text))
    
    def predict_with_merge(self, text: Union[str, List[str]]) -> Union[LanguageToolRemoteResult, List[LanguageToolRemoteResult]]:
        """
        Run all clients, merge their results, and return merged result(s).
        
        Args:
            text: Input text or list of texts to process
            
        Returns:
            Merged LanguageToolRemoteResult (single text) or List of merged results (batch)
        """
        # Handle batch processing
        if isinstance(text, list):
            try:
                loop = asyncio.get_running_loop()
                # If loop is already running, run in new thread
                with ThreadPoolExecutor(max_workers=1) as executor:
                    future = executor.submit(asyncio.run, self._predict_batch_with_merge_async(text))
                    return future.result()
            except RuntimeError:
                # No event loop exists, create one
                return asyncio.run(self._predict_batch_with_merge_async(text))
        
        # Handle single text processing
        try:
            loop = asyncio.get_running_loop()
            # If loop is already running, run in new thread
            with ThreadPoolExecutor(max_workers=1) as executor:
                future = executor.submit(asyncio.run, self._predict_with_merge_async(text))
                return future.result()
        except RuntimeError:
            # No event loop exists, create one
            return asyncio.run(self._predict_with_merge_async(text))
    
    def predict_batch(self, texts: List[str]) -> List[List[LanguageToolRemoteResult]]:
        """
        Process multiple texts concurrently across all clients.
        
        Args:
            texts: List of input texts to process
            
        Returns:
            List of results for each text, where each result is a list from all clients
        """
        try:
            loop = asyncio.get_running_loop()
            # If loop is already running, run in new thread
            with ThreadPoolExecutor(max_workers=1) as executor:
                future = executor.submit(asyncio.run, self._predict_batch_async(texts))
                return future.result()
        except RuntimeError:
            # No event loop exists, create one
            return asyncio.run(self._predict_batch_async(texts))
    
    def __call__(self, text: str) -> List[LanguageToolRemoteResult]:
        """Make the client callable."""
        return self.predict(text)
    
    def close(self):
        """Cleanup resources."""
        if hasattr(self, '_executor'):
            self._executor.shutdown(wait=True)
    
    def __enter__(self):
        """Context manager entry."""
        return self
    
    def __exit__(self, exc_type, exc_val, exc_tb):
        """Context manager exit."""
        self.close()
        return False
    
    async def __aenter__(self):
        """Async context manager entry."""
        return self
    
    async def __aexit__(self, exc_type, exc_val, exc_tb):
        """Async context manager exit."""
        self.close()
        return False



# from .triton.triton_templates import
import json
from jinja2 import Environment, FileSystemLoader
from pathlib import Path
import yaml
from grammared_language.utils .config_parser import BaseModelConfig, ModelsConfig
from grammared_language.utils import config_parser

DEFAULT_TEMPLATE_FOLDER = str(Path(__file__).parent / "triton_templates")

SUPPORTED_MODEL_TYPES = ("gector", "grammared_classifier", "coedit")

TEMPLATE_FILES_BY_MODEL_TYPE = {
    "gector": {
        "config": "gector.config.pbtxt.jinja",
        "model": "gector.model.py.jinja"
    },
    "grammared_classifier": {
        "config": "grammared_classifier.config.pbtxt.jinja",
        "model": "grammared_classifier.model.py.jinja"
    },
    "coedit": {
        "config": "text2text.config.pbtxt.jinja",
        "model": "text2text.model.py.jinja"
    }
}

class TritonRepoBuilder:
    def __init__(self, template_folder: str|None=None):
        if template_folder is None:
            template_folder = DEFAULT_TEMPLATE_FOLDER
        self.template_folder = template_folder
        self.jina_loader = FileSystemLoader(searchpath=self.template_folder)
        self.jinja_env = Environment(loader=self.jina_loader)

    def build_model_repo(self, repo_folder: str, config: ModelsConfig|None=None, config_path:str|None=None):
        
        if not (config or config_path):
            raise ValueError("Either config or config_path must be provided.")
        
        if config is None:
            config = config_parser.load_config_from_file(config_path)

        for m in config.models:
            model_config = config.models[m]
            if model_config.type not in SUPPORTED_MODEL_TYPES:
                raise ValueError(f"Unsupported model type: {model_config.type}")

            name = m
            if model_config.serving_config.triton_model_name is not None:
                name = model_config.serving_config.triton_model_name

            self._build_model_repo(
                    name=name,
                    model_config=model_config,
                    repo_folder=repo_folder
            )
            
    def _build_model_repo(self, name:str, model_config:BaseModelConfig, repo_folder:str, model_version:int=1):
        model_type = model_config.type
        pretrained_model_name_or_path = model_config.serving_config.pretrained_model_name_or_path
        template_files = TEMPLATE_FILES_BY_MODEL_TYPE[model_type]
        output_folder = Path(repo_folder) / name
        model_folder = output_folder / str(model_version)
        config_file = self.jinja_env.get_template(template_files["config"]).render({
            "model_name": name,
            "pretrained_model_name_or_path": pretrained_model_name_or_path,
            "json_model_config": json.dumps(model_config.model_dump_json())[1:-1],
        })

        model_file = self.jinja_env.get_template(template_files["model"]).render()
        Path(output_folder).mkdir(parents=True, exist_ok=True)
        Path(model_folder).mkdir(parents=True, exist_ok=True)
        with open(output_folder / "config.pbtxt", "w") as f:
            f.write(config_file)

        with open(model_folder / "model.py", "w") as f:
            f.write(model_file)


"""Triton Inference Server model implementations."""

import importlib
import pkgutil
from pathlib import Path

try:
    import triton_python_backend_utils as pb_utils
    is_triton_server = True
except ImportError:
    is_triton_server = False


# Programmatically discover and import all modules
__all__ = []
_module_path = Path(__file__).parent

for _, module_name, _ in pkgutil.iter_modules([str(_module_path)]):
    if not module_name.startswith('_'):
        if module_name.startswith('triton_') and not is_triton_server:
            # skips Triton-specific models when not in Triton server environment
            continue
        module = importlib.import_module(f'.{module_name}', package=__name__)
        # Export all public attributes from the module
        for attr_name in dir(module):
            if not attr_name.startswith('_'):
                attr = getattr(module, attr_name)
                # Only export classes and functions defined in this module
                if hasattr(attr, '__module__') and attr.__module__.startswith(__name__):
                    globals()[attr_name] = attr
                    __all__.append(attr_name)



# Copyright 2023, NVIDIA CORPORATION & AFFILIATES. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#  * Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#  * Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#  * Neither the name of NVIDIA CORPORATION nor the names of its
#    contributors may be used to endorse or promote products derived
#    from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
# EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
# OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
from ast import Load
import logging
import os
import json
from xml.parsers.expat import model

import numpy as np
import torch
import transformers
import triton_python_backend_utils as pb_utils
from grammared_language.utils.config_parser import get_model_config, BaseModelConfig


logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.DEBUG) 


DEFAULT_MODEL_BACKEND = "transformers"
def load_pipeline_from_config(model_config:BaseModelConfig, task="text2text-generation", backend:str="transformers", fallback:bool=True, **kwargs):
    """Load a HuggingFace model based on the provided configuration.

    Args:
        model_config (BaseModelConfig): Model configuration

    Returns:
        model: Loaded HuggingFace model.
    """
    # Extract model parameters from configuration
    default_hf_model = "HuggingFaceTB/SmolLM2-135M"
    if model_config.serving_config.pretrained_model_name_or_path:
        hf_model = model_config.serving_config.pretrained_model_name_or_path
    else:
        hf_model = default_hf_model
    logger.warning(f"Loading model: {hf_model} with backend: {backend}")
    try:
        if backend == "ort":
            from optimum import onnxruntime
            # pipeline = onnxruntime.pipeline(task=task, model=hf_model, **kwargs)
            tokenizer = transformers.AutoTokenizer.from_pretrained(hf_model, truncation=True)
            ORTModelClass = onnxruntime.pipelines.ORT_TASKS_MAPPING[task][0]
            provider = "CPUExecutionProvider"
            if 'cuda' in kwargs['device'].lower():
                provider = 'CUDAExecutionProvider'
            model = ORTModelClass.from_pretrained(
                hf_model,
                provider=provider
            )
            logger.warning(f"model_config.model_config: {model_config.model_config}")
            logger.warning(f"kwargs: {kwargs}")
            pipeline = transformers.pipeline(
                task, 
                model=model, 
                tokenizer=tokenizer,
                **model_config.model_init_config.model_dump(),
                **kwargs)
        else:
            pipeline = transformers.pipeline(
                task=task, model=hf_model, 
                **model_config.model_init_config.model_dump(),
                **kwargs
            )
    except Exception as e:
        if fallback and backend != DEFAULT_MODEL_BACKEND:
            logger.warning(f"Failed to load model {hf_model}: {str(e)} using backend: {backend}. Falling back to default model backend {DEFAULT_MODEL_BACKEND}.")
            pipeline = transformers.pipeline(
                task=task, model=hf_model, 
                **model_config.model_init_config.model_dump(),
                **kwargs
            )
        else:
            raise RuntimeError(f"Failed to load model {hf_model}: {str(e)}")
    return pipeline


class TritonTransformersPythonModel:
    PIPELINE_TASK = "text-generation"
    DEFAULT_HF_MODEL = "HuggingFaceTB/SmolLM2-135M"
    def initialize(self, args):
        self.logger = pb_utils.Logger
        self.model_config = json.loads(args["model_config"])
        self.model_params = self.model_config.get("parameters", {})
        self.grammared_language_model_config = json.loads(
            self.model_config.get('parameters', {}).get('grammared_language_model_config', {}).get('string_value', "{}")
        )
        self.grammared_language_model_config = get_model_config(self.model_config.get("name", "transfomers_model"), self.grammared_language_model_config)

        default_max_gen_length = 30
        self.max_output_length = default_max_gen_length
        if hasattr(self.grammared_language_model_config.model_inference_config, 'max_length'):
            self.max_output_length = self.grammared_language_model_config.model_inference_config.max_length
        # Check for user-specified model name in model config parameters
        hf_model = self.model_params.get("pretrained_model_name_or_path", {}).get(
            "string_value", self.DEFAULT_HF_MODEL
        )

        logger.warning(f"Loading HuggingFace model: {hf_model}...")
        logger.warning(f"Loading grammared_language_model_config: {self.grammared_language_model_config}...")

        # Get model instance device configuration
        model_device_type = args['model_instance_kind']
        model_instance_device_id = args['model_instance_device_id']
        if self.grammared_language_model_config.serving_config.device == 'cpu':
            self.device = 'cpu'
        elif self.grammared_language_model_config.serving_config.device == 'cuda':
            self.device = f'cuda:{model_instance_device_id}'
        else:
            # mimic 'auto' behavior
            if torch.cuda.is_available():
                self.device = f'cuda:{model_instance_device_id}'
            else:
                self.device = 'cpu'

        # # Assume tokenizer available for same model
        # self.tokenizer = transformers.AutoTokenizer.from_pretrained(hf_model)
        # self.pipeline = transformers.pipeline(
        #     self.PIPELINE_TASK,
        #     model=hf_model,
        #     # torch_dtype=torch.float16,
        #     tokenizer=self.tokenizer,
        #     device_map="auto",
        # )

        self.pipeline = load_pipeline_from_config(
            self.grammared_language_model_config,
            task=self.PIPELINE_TASK,
            backend=self.grammared_language_model_config.serving_config.backend or DEFAULT_MODEL_BACKEND,
            fallback=True,
            device=self.device,
        )

    def execute(self, requests):
        responses = []
        for request in requests:
            # Get input tensor with shape [batch_size, 1]
            input_tensor = pb_utils.get_input_tensor_by_name(request, "text_input")
            input_data = input_tensor.as_numpy()
            
            # Process all items in the batch
            prompts = []
            for i in range(len(input_data)):
                prompt = input_data[i][0].decode("utf-8")
                prompts.append(prompt)
                logger.warning(f"Batch item {i}: {prompt}")
            
            response = self.generate_batch(prompts)
            responses.append(response)

        return responses

    def generate_batch(self, prompts):
        """Generate text for a batch of prompts."""
        sequences = self.pipeline(
            prompts,
            max_length=self.max_output_length,
        )
        all_texts = [seq["generated_text"] for seq in sequences]
        logger.warning(f"Generated: {all_texts}")
        # Return batch output with shape [batch_size, -1]
        tensor = pb_utils.Tensor("text_output", np.array(all_texts, dtype=np.object_).reshape(-1, 1))
        output_tensors = [tensor]
        response = pb_utils.InferenceResponse(output_tensors=output_tensors)
        return response

    def finalize(self):
        print("Cleaning up...")


import threading
import os

import os
import threading
from collections import OrderedDict
from grammared_language.language_tool.output_models import LanguageToolRemoteResult
from .grpc_gen import ml_server_pb2

class SimpleCacheStore:
    """Thread-safe in-memory cache store."""
    def __init__(self, max_size=10000):
        self.store = OrderedDict()
        self.max_size = max_size
        self._lock = threading.RLock()
    
    def add(self, text: str, result: LanguageToolRemoteResult):
        with self._lock:
            if len(self.store) >= self.max_size:
                self.store.popitem(last=False)
            self.store[text] = result
    
    def contains(self, text: str) -> bool:
        with self._lock:
            return text in self.store
    
    def get(self, text: str):
        with self._lock:
            self.store.move_to_end(text)
            return self.store.get(text)


#!/usr/bin/env python
"""
Script to regenerate gRPC code from ml_server.proto.

Usage:
    python generate_grpc.py

This regenerates the protobuf and gRPC Python code from the LanguageTool
ml_server.proto definition.
"""

import subprocess
import sys
from pathlib import Path
import re

def fix_relative_imports(grpc_gen_dir):
    """Convert absolute imports to relative imports in generated files."""
    py_files = list(grpc_gen_dir.glob("*.py"))
    
    for file_path in py_files:
        content = file_path.read_text()
        # Replace "import ml_server_pb2" with "from . import ml_server_pb2"
        updated_content = re.sub(
            r'^import ml_server_pb2 as',
            'from . import ml_server_pb2 as',
            content,
            flags=re.MULTILINE
        )
        if updated_content != content:
            file_path.write_text(updated_content)
            print(f"✓ Fixed imports in {file_path.name}")

def regenerate_grpc():
    """Regenerate gRPC code from ml_server.proto."""
    src_dir = Path(__file__).parent.parent / "language_tool"
    grpc_gen_dir = src_dir / "grpc_gen"
    proto_file = src_dir / "ml_server.proto"
    
    if not proto_file.exists():
        print(f"Error: {proto_file} not found")
        return False
    
    if not grpc_gen_dir.exists():
        grpc_gen_dir.mkdir(parents=True, exist_ok=True)
        print(f"Created directory: {grpc_gen_dir}")
    
    print(f"Regenerating gRPC code from {proto_file}...")
    print(f"Output directory: {grpc_gen_dir}")
    
    cmd = [
        sys.executable, "-m", "grpc_tools.protoc",
        f"-I{src_dir}",
        f"--python_out={grpc_gen_dir}",
        f"--pyi_out={grpc_gen_dir}",
        f"--grpc_python_out={grpc_gen_dir}",
        str(proto_file)
    ]
    
    try:
        result = subprocess.run(cmd, check=True, capture_output=True, text=True)
        print("✓ gRPC code regenerated successfully")
        
        # Fix relative imports
        fix_relative_imports(grpc_gen_dir)
                
        if result.stdout:
            print(result.stdout)
        return True
    except subprocess.CalledProcessError as e:
        print(f"✗ Error regenerating gRPC code:")
        print(f"Command: {' '.join(cmd)}")
        if e.stdout:
            print("STDOUT:", e.stdout)
        if e.stderr:
            print("STDERR:", e.stderr)
        return False
    except FileNotFoundError:
        print("✗ grpc_tools.protoc not found. Install with: pip install grpcio-tools")
        return False

if __name__ == "__main__":
    success = regenerate_grpc()
    sys.exit(0 if success else 1)



from google.protobuf.internal import containers as _containers
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class AnalyzeRequest(_message.Message):
    __slots__ = ("text", "options")
    TEXT_FIELD_NUMBER: _ClassVar[int]
    OPTIONS_FIELD_NUMBER: _ClassVar[int]
    text: str
    options: ProcessingOptions
    def __init__(self, text: _Optional[str] = ..., options: _Optional[_Union[ProcessingOptions, _Mapping]] = ...) -> None: ...

class ProcessingOptions(_message.Message):
    __slots__ = ("language", "tempOff", "level", "premium", "enabledOnly", "enabledRules", "disabledRules")
    class Level(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
        __slots__ = ()
        picky: _ClassVar[ProcessingOptions.Level]
        academic: _ClassVar[ProcessingOptions.Level]
        clarity: _ClassVar[ProcessingOptions.Level]
        professional: _ClassVar[ProcessingOptions.Level]
        creative: _ClassVar[ProcessingOptions.Level]
        customer: _ClassVar[ProcessingOptions.Level]
        jobapp: _ClassVar[ProcessingOptions.Level]
        objective: _ClassVar[ProcessingOptions.Level]
        elegant: _ClassVar[ProcessingOptions.Level]
        defaultLevel: _ClassVar[ProcessingOptions.Level]
    picky: ProcessingOptions.Level
    academic: ProcessingOptions.Level
    clarity: ProcessingOptions.Level
    professional: ProcessingOptions.Level
    creative: ProcessingOptions.Level
    customer: ProcessingOptions.Level
    jobapp: ProcessingOptions.Level
    objective: ProcessingOptions.Level
    elegant: ProcessingOptions.Level
    defaultLevel: ProcessingOptions.Level
    LANGUAGE_FIELD_NUMBER: _ClassVar[int]
    TEMPOFF_FIELD_NUMBER: _ClassVar[int]
    LEVEL_FIELD_NUMBER: _ClassVar[int]
    PREMIUM_FIELD_NUMBER: _ClassVar[int]
    ENABLEDONLY_FIELD_NUMBER: _ClassVar[int]
    ENABLEDRULES_FIELD_NUMBER: _ClassVar[int]
    DISABLEDRULES_FIELD_NUMBER: _ClassVar[int]
    language: str
    tempOff: bool
    level: ProcessingOptions.Level
    premium: bool
    enabledOnly: bool
    enabledRules: _containers.RepeatedScalarFieldContainer[str]
    disabledRules: _containers.RepeatedScalarFieldContainer[str]
    def __init__(self, language: _Optional[str] = ..., tempOff: bool = ..., level: _Optional[_Union[ProcessingOptions.Level, str]] = ..., premium: bool = ..., enabledOnly: bool = ..., enabledRules: _Optional[_Iterable[str]] = ..., disabledRules: _Optional[_Iterable[str]] = ...) -> None: ...

class AnalyzeResponse(_message.Message):
    __slots__ = ("sentences",)
    SENTENCES_FIELD_NUMBER: _ClassVar[int]
    sentences: _containers.RepeatedCompositeFieldContainer[AnalyzedSentence]
    def __init__(self, sentences: _Optional[_Iterable[_Union[AnalyzedSentence, _Mapping]]] = ...) -> None: ...

class ProcessRequest(_message.Message):
    __slots__ = ("sentences", "options")
    SENTENCES_FIELD_NUMBER: _ClassVar[int]
    OPTIONS_FIELD_NUMBER: _ClassVar[int]
    sentences: _containers.RepeatedCompositeFieldContainer[AnalyzedSentence]
    options: ProcessingOptions
    def __init__(self, sentences: _Optional[_Iterable[_Union[AnalyzedSentence, _Mapping]]] = ..., options: _Optional[_Union[ProcessingOptions, _Mapping]] = ...) -> None: ...

class ProcessResponse(_message.Message):
    __slots__ = ("rawMatches", "matches")
    RAWMATCHES_FIELD_NUMBER: _ClassVar[int]
    MATCHES_FIELD_NUMBER: _ClassVar[int]
    rawMatches: _containers.RepeatedCompositeFieldContainer[Match]
    matches: _containers.RepeatedCompositeFieldContainer[Match]
    def __init__(self, rawMatches: _Optional[_Iterable[_Union[Match, _Mapping]]] = ..., matches: _Optional[_Iterable[_Union[Match, _Mapping]]] = ...) -> None: ...

class AnalyzedMatchRequest(_message.Message):
    __slots__ = ("sentences", "inputLogging", "textSessionID")
    SENTENCES_FIELD_NUMBER: _ClassVar[int]
    INPUTLOGGING_FIELD_NUMBER: _ClassVar[int]
    TEXTSESSIONID_FIELD_NUMBER: _ClassVar[int]
    sentences: _containers.RepeatedCompositeFieldContainer[AnalyzedSentence]
    inputLogging: bool
    textSessionID: _containers.RepeatedScalarFieldContainer[int]
    def __init__(self, sentences: _Optional[_Iterable[_Union[AnalyzedSentence, _Mapping]]] = ..., inputLogging: bool = ..., textSessionID: _Optional[_Iterable[int]] = ...) -> None: ...

class AnalyzedSentence(_message.Message):
    __slots__ = ("text", "tokens")
    TEXT_FIELD_NUMBER: _ClassVar[int]
    TOKENS_FIELD_NUMBER: _ClassVar[int]
    text: str
    tokens: _containers.RepeatedCompositeFieldContainer[AnalyzedTokenReadings]
    def __init__(self, text: _Optional[str] = ..., tokens: _Optional[_Iterable[_Union[AnalyzedTokenReadings, _Mapping]]] = ...) -> None: ...

class AnalyzedTokenReadings(_message.Message):
    __slots__ = ("readings", "chunkTags", "startPos")
    READINGS_FIELD_NUMBER: _ClassVar[int]
    CHUNKTAGS_FIELD_NUMBER: _ClassVar[int]
    STARTPOS_FIELD_NUMBER: _ClassVar[int]
    readings: _containers.RepeatedCompositeFieldContainer[AnalyzedToken]
    chunkTags: _containers.RepeatedScalarFieldContainer[str]
    startPos: int
    def __init__(self, readings: _Optional[_Iterable[_Union[AnalyzedToken, _Mapping]]] = ..., chunkTags: _Optional[_Iterable[str]] = ..., startPos: _Optional[int] = ...) -> None: ...

class AnalyzedToken(_message.Message):
    __slots__ = ("token", "posTag", "lemma")
    TOKEN_FIELD_NUMBER: _ClassVar[int]
    POSTAG_FIELD_NUMBER: _ClassVar[int]
    LEMMA_FIELD_NUMBER: _ClassVar[int]
    token: str
    posTag: str
    lemma: str
    def __init__(self, token: _Optional[str] = ..., posTag: _Optional[str] = ..., lemma: _Optional[str] = ...) -> None: ...

class PostProcessingRequest(_message.Message):
    __slots__ = ("sentences", "matches", "inputLogging", "textSessionID")
    SENTENCES_FIELD_NUMBER: _ClassVar[int]
    MATCHES_FIELD_NUMBER: _ClassVar[int]
    INPUTLOGGING_FIELD_NUMBER: _ClassVar[int]
    TEXTSESSIONID_FIELD_NUMBER: _ClassVar[int]
    sentences: _containers.RepeatedScalarFieldContainer[str]
    matches: _containers.RepeatedCompositeFieldContainer[MatchList]
    inputLogging: bool
    textSessionID: _containers.RepeatedScalarFieldContainer[int]
    def __init__(self, sentences: _Optional[_Iterable[str]] = ..., matches: _Optional[_Iterable[_Union[MatchList, _Mapping]]] = ..., inputLogging: bool = ..., textSessionID: _Optional[_Iterable[int]] = ...) -> None: ...

class MatchRequest(_message.Message):
    __slots__ = ("sentences", "inputLogging", "textSessionID")
    SENTENCES_FIELD_NUMBER: _ClassVar[int]
    INPUTLOGGING_FIELD_NUMBER: _ClassVar[int]
    TEXTSESSIONID_FIELD_NUMBER: _ClassVar[int]
    sentences: _containers.RepeatedScalarFieldContainer[str]
    inputLogging: bool
    textSessionID: _containers.RepeatedScalarFieldContainer[int]
    def __init__(self, sentences: _Optional[_Iterable[str]] = ..., inputLogging: bool = ..., textSessionID: _Optional[_Iterable[int]] = ...) -> None: ...

class MatchResponse(_message.Message):
    __slots__ = ("sentenceMatches",)
    SENTENCEMATCHES_FIELD_NUMBER: _ClassVar[int]
    sentenceMatches: _containers.RepeatedCompositeFieldContainer[MatchList]
    def __init__(self, sentenceMatches: _Optional[_Iterable[_Union[MatchList, _Mapping]]] = ...) -> None: ...

class MatchList(_message.Message):
    __slots__ = ("matches",)
    MATCHES_FIELD_NUMBER: _ClassVar[int]
    matches: _containers.RepeatedCompositeFieldContainer[Match]
    def __init__(self, matches: _Optional[_Iterable[_Union[Match, _Mapping]]] = ...) -> None: ...

class Match(_message.Message):
    __slots__ = ("offset", "length", "id", "sub_id", "suggestions", "ruleDescription", "matchDescription", "matchShortDescription", "url", "suggestedReplacements", "autoCorrect", "type", "contextForSureMatch", "rule")
    class MatchType(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
        __slots__ = ()
        UnknownWord: _ClassVar[Match.MatchType]
        Hint: _ClassVar[Match.MatchType]
        Other: _ClassVar[Match.MatchType]
    UnknownWord: Match.MatchType
    Hint: Match.MatchType
    Other: Match.MatchType
    OFFSET_FIELD_NUMBER: _ClassVar[int]
    LENGTH_FIELD_NUMBER: _ClassVar[int]
    ID_FIELD_NUMBER: _ClassVar[int]
    SUB_ID_FIELD_NUMBER: _ClassVar[int]
    SUGGESTIONS_FIELD_NUMBER: _ClassVar[int]
    RULEDESCRIPTION_FIELD_NUMBER: _ClassVar[int]
    MATCHDESCRIPTION_FIELD_NUMBER: _ClassVar[int]
    MATCHSHORTDESCRIPTION_FIELD_NUMBER: _ClassVar[int]
    URL_FIELD_NUMBER: _ClassVar[int]
    SUGGESTEDREPLACEMENTS_FIELD_NUMBER: _ClassVar[int]
    AUTOCORRECT_FIELD_NUMBER: _ClassVar[int]
    TYPE_FIELD_NUMBER: _ClassVar[int]
    CONTEXTFORSUREMATCH_FIELD_NUMBER: _ClassVar[int]
    RULE_FIELD_NUMBER: _ClassVar[int]
    offset: int
    length: int
    id: str
    sub_id: str
    suggestions: _containers.RepeatedScalarFieldContainer[str]
    ruleDescription: str
    matchDescription: str
    matchShortDescription: str
    url: str
    suggestedReplacements: _containers.RepeatedCompositeFieldContainer[SuggestedReplacement]
    autoCorrect: bool
    type: Match.MatchType
    contextForSureMatch: int
    rule: Rule
    def __init__(self, offset: _Optional[int] = ..., length: _Optional[int] = ..., id: _Optional[str] = ..., sub_id: _Optional[str] = ..., suggestions: _Optional[_Iterable[str]] = ..., ruleDescription: _Optional[str] = ..., matchDescription: _Optional[str] = ..., matchShortDescription: _Optional[str] = ..., url: _Optional[str] = ..., suggestedReplacements: _Optional[_Iterable[_Union[SuggestedReplacement, _Mapping]]] = ..., autoCorrect: bool = ..., type: _Optional[_Union[Match.MatchType, str]] = ..., contextForSureMatch: _Optional[int] = ..., rule: _Optional[_Union[Rule, _Mapping]] = ...) -> None: ...

class Rule(_message.Message):
    __slots__ = ("sourceFile", "issueType", "tempOff", "category", "isPremium", "tags")
    class Tag(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
        __slots__ = ()
        picky: _ClassVar[Rule.Tag]
        academic: _ClassVar[Rule.Tag]
        clarity: _ClassVar[Rule.Tag]
        professional: _ClassVar[Rule.Tag]
        creative: _ClassVar[Rule.Tag]
        customer: _ClassVar[Rule.Tag]
        jobapp: _ClassVar[Rule.Tag]
        objective: _ClassVar[Rule.Tag]
        elegant: _ClassVar[Rule.Tag]
    picky: Rule.Tag
    academic: Rule.Tag
    clarity: Rule.Tag
    professional: Rule.Tag
    creative: Rule.Tag
    customer: Rule.Tag
    jobapp: Rule.Tag
    objective: Rule.Tag
    elegant: Rule.Tag
    SOURCEFILE_FIELD_NUMBER: _ClassVar[int]
    ISSUETYPE_FIELD_NUMBER: _ClassVar[int]
    TEMPOFF_FIELD_NUMBER: _ClassVar[int]
    CATEGORY_FIELD_NUMBER: _ClassVar[int]
    ISPREMIUM_FIELD_NUMBER: _ClassVar[int]
    TAGS_FIELD_NUMBER: _ClassVar[int]
    sourceFile: str
    issueType: str
    tempOff: bool
    category: RuleCategory
    isPremium: bool
    tags: _containers.RepeatedScalarFieldContainer[Rule.Tag]
    def __init__(self, sourceFile: _Optional[str] = ..., issueType: _Optional[str] = ..., tempOff: bool = ..., category: _Optional[_Union[RuleCategory, _Mapping]] = ..., isPremium: bool = ..., tags: _Optional[_Iterable[_Union[Rule.Tag, str]]] = ...) -> None: ...

class RuleCategory(_message.Message):
    __slots__ = ("id", "name")
    ID_FIELD_NUMBER: _ClassVar[int]
    NAME_FIELD_NUMBER: _ClassVar[int]
    id: str
    name: str
    def __init__(self, id: _Optional[str] = ..., name: _Optional[str] = ...) -> None: ...

class SuggestedReplacement(_message.Message):
    __slots__ = ("replacement", "description", "suffix", "confidence", "type")
    class SuggestionType(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
        __slots__ = ()
        Default: _ClassVar[SuggestedReplacement.SuggestionType]
        Translation: _ClassVar[SuggestedReplacement.SuggestionType]
        Curated: _ClassVar[SuggestedReplacement.SuggestionType]
    Default: SuggestedReplacement.SuggestionType
    Translation: SuggestedReplacement.SuggestionType
    Curated: SuggestedReplacement.SuggestionType
    REPLACEMENT_FIELD_NUMBER: _ClassVar[int]
    DESCRIPTION_FIELD_NUMBER: _ClassVar[int]
    SUFFIX_FIELD_NUMBER: _ClassVar[int]
    CONFIDENCE_FIELD_NUMBER: _ClassVar[int]
    TYPE_FIELD_NUMBER: _ClassVar[int]
    replacement: str
    description: str
    suffix: str
    confidence: float
    type: SuggestedReplacement.SuggestionType
    def __init__(self, replacement: _Optional[str] = ..., description: _Optional[str] = ..., suffix: _Optional[str] = ..., confidence: _Optional[float] = ..., type: _Optional[_Union[SuggestedReplacement.SuggestionType, str]] = ...) -> None: ...



"""
GECToR (Grammatical Error Correction: Tag, Not Rewrite) Model for Triton Inference Server.

This model uses the HuggingFace transformers library to serve the
gotutiyan/gector-bert-base-cased-5k model for grammar error correction.
"""
import logging
import json
import numpy as np
import triton_python_backend_utils as pb_utils
from transformers import AutoModel
import torch
from gector import GECToR
from grammared_language.utils.config_parser import get_model_config

logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.DEBUG)  # Set default level to WARNING


class TritonGectorPythonModel:
    """Python model for GECToR grammar error correction."""

    def initialize(self, args):
        """
        Initialize the model.

        Args:
            args: Dictionary containing model configuration
        """
        self.model_config = json.loads(args['model_config'])
        self.grammared_language_model_config = json.loads(
            self.model_config.get('parameters', {}).get('grammared_language_model_config', {}).get('string_value', "{}")
        )
        self.grammared_language_model_config = get_model_config(self.model_config.get("name", "gector_model"), self.grammared_language_model_config)
        logger.warning(f"Loaded model config: {self.model_config}")
        logger.warning(f"Loading grammared_language_model_config: {self.grammared_language_model_config}...")
        # Get model instance device configuration

        model_device_type = args['model_instance_kind']
        model_instance_device_id = args['model_instance_device_id']

        if self.grammared_language_model_config.serving_config.device == 'cpu':
            self.device = 'cpu'
        elif self.grammared_language_model_config.serving_config.device == 'cuda':
            self.device = f'cuda:{model_instance_device_id}'
        else:
            # mimic 'auto' behavior
            if torch.cuda.is_available():
                self.device = f'cuda:{model_instance_device_id}'
            else:
                self.device = 'cpu'

        # Initialize attributes for cleanup
        self.model = None

        # Load the HuggingFace model
        if 'pretrained_model_name_or_path' in self.model_config['parameters']:
            model_name = self.model_config['parameters']['pretrained_model_name_or_path']['string_value']
        else:
            raise RuntimeError(f"Failed to get model name from configuration: {self.model_config}")

        try:
            self.model = GECToR.from_pretrained(model_name, device_map=self.device)

            logger.warning(f"Loaded model {model_name}")
            # logger.warning(f"Model config: {self.model.config}")
            # self.model.to(self.device)
            # first_encoder_layer = self.model.encoder.layer[0]
            # query_weights = first_encoder_layer.attention.self.query.weight
            # logger.warning(f"First encoder layer query weights: {query_weights}")
            # self.model.eval()


        except Exception as e:
            raise RuntimeError(f"Failed to load model {model_name}: {str(e)}")

    def execute(self, requests):
        """
        Execute inference on a batch of requests.

        Args:
            requests: List of pb_utils.InferenceRequest

        Returns:
            List of pb_utils.InferenceResponse
        """
        responses = []

        for request in requests:
            try:
                # Get input tensors
                input_ids_tensor = pb_utils.get_input_tensor_by_name(request, "input_ids")
                attention_mask_tensor = pb_utils.get_input_tensor_by_name(request, "attention_mask")

                # Convert to numpy arrays
                input_ids_np = input_ids_tensor.as_numpy()
                attention_mask_np = attention_mask_tensor.as_numpy()

                # Convert to torch tensors and move to device
                input_ids = torch.from_numpy(input_ids_np).to(self.device)
                attention_mask = torch.from_numpy(attention_mask_np).to(self.device)

                # Run inference
                gector_output = self.model.forward(
                    input_ids=input_ids,
                    attention_mask=attention_mask
                )

                # Get logits
                logits_d = gector_output.logits_d
                logits_labels = gector_output.logits_labels

                # Convert logits to numpy
                logits_labels_np = logits_labels.cpu().detach().numpy().astype(np.float32)
                logits_d_np = logits_d.cpu().detach().numpy().astype(np.float32)
                # logger.warning(f"logits_labels_np: {logits_labels_np}")
                output_logits_labels = pb_utils.Tensor(
                    "logits_labels",
                    logits_labels_np
                )

                output_logits_d = pb_utils.Tensor(
                    "logits_d",
                    logits_d_np
                )

                # logger.warning(f"output_logits_d: {output_logits_d}")
                # Create inference response
                response = pb_utils.InferenceResponse(
                    output_tensors=[output_logits_labels, output_logits_d]
                )

            except Exception as e:
                # Return error response
                error_msg = f"Error during inference: {str(e)}"
                response = pb_utils.InferenceResponse(
                    output_tensors=[],
                    error=pb_utils.TritonError(error_msg)
                )

            responses.append(response)

        return responses

    def finalize(self):
        """Clean up resources."""
        # Clean up model safely
        if hasattr(self, 'model') and self.model is not None:
            del self.model
        if torch.cuda.is_available():
            torch.cuda.empty_cache()



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: ml_server.proto
# Protobuf Python Version: 5.27.2
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    5,
    27,
    2,
    '',
    'ml_server.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x0fml_server.proto\x12\x0clt_ml_server\"P\n\x0e\x41nalyzeRequest\x12\x0c\n\x04text\x18\x01 \x01(\t\x12\x30\n\x07options\x18\x02 \x01(\x0b\x32\x1f.lt_ml_server.ProcessingOptions\"\xd7\x02\n\x11ProcessingOptions\x12\x10\n\x08language\x18\x01 \x01(\t\x12\x0f\n\x07tempOff\x18\x02 \x01(\x08\x12\x34\n\x05level\x18\x03 \x01(\x0e\x32%.lt_ml_server.ProcessingOptions.Level\x12\x0f\n\x07premium\x18\x04 \x01(\x08\x12\x13\n\x0b\x65nabledOnly\x18\x05 \x01(\x08\x12\x14\n\x0c\x65nabledRules\x18\x06 \x03(\t\x12\x15\n\rdisabledRules\x18\x07 \x03(\t\"\x95\x01\n\x05Level\x12\t\n\x05picky\x10\x00\x12\x0c\n\x08\x61\x63\x61\x64\x65mic\x10\x01\x12\x0b\n\x07\x63larity\x10\x02\x12\x10\n\x0cprofessional\x10\x03\x12\x0c\n\x08\x63reative\x10\x04\x12\x0c\n\x08\x63ustomer\x10\x05\x12\n\n\x06jobapp\x10\x06\x12\r\n\tobjective\x10\x07\x12\x0b\n\x07\x65legant\x10\x08\x12\x10\n\x0c\x64\x65\x66\x61ultLevel\x10\t\"D\n\x0f\x41nalyzeResponse\x12\x31\n\tsentences\x18\x01 \x03(\x0b\x32\x1e.lt_ml_server.AnalyzedSentence\"u\n\x0eProcessRequest\x12\x31\n\tsentences\x18\x01 \x03(\x0b\x32\x1e.lt_ml_server.AnalyzedSentence\x12\x30\n\x07options\x18\x02 \x01(\x0b\x32\x1f.lt_ml_server.ProcessingOptions\"`\n\x0fProcessResponse\x12\'\n\nrawMatches\x18\x01 \x03(\x0b\x32\x13.lt_ml_server.Match\x12$\n\x07matches\x18\x02 \x03(\x0b\x32\x13.lt_ml_server.Match\"v\n\x14\x41nalyzedMatchRequest\x12\x31\n\tsentences\x18\x01 \x03(\x0b\x32\x1e.lt_ml_server.AnalyzedSentence\x12\x14\n\x0cinputLogging\x18\x02 \x01(\x08\x12\x15\n\rtextSessionID\x18\x03 \x03(\x03\"U\n\x10\x41nalyzedSentence\x12\x0c\n\x04text\x18\x01 \x01(\t\x12\x33\n\x06tokens\x18\x02 \x03(\x0b\x32#.lt_ml_server.AnalyzedTokenReadings\"k\n\x15\x41nalyzedTokenReadings\x12-\n\x08readings\x18\x01 \x03(\x0b\x32\x1b.lt_ml_server.AnalyzedToken\x12\x11\n\tchunkTags\x18\x02 \x03(\t\x12\x10\n\x08startPos\x18\x03 \x01(\x05\"=\n\rAnalyzedToken\x12\r\n\x05token\x18\x01 \x01(\t\x12\x0e\n\x06posTag\x18\x02 \x01(\t\x12\r\n\x05lemma\x18\x03 \x01(\t\"\x81\x01\n\x15PostProcessingRequest\x12\x11\n\tsentences\x18\x01 \x03(\t\x12(\n\x07matches\x18\x02 \x03(\x0b\x32\x17.lt_ml_server.MatchList\x12\x14\n\x0cinputLogging\x18\x03 \x01(\x08\x12\x15\n\rtextSessionID\x18\x04 \x03(\x03\"N\n\x0cMatchRequest\x12\x11\n\tsentences\x18\x01 \x03(\t\x12\x14\n\x0cinputLogging\x18\x02 \x01(\x08\x12\x15\n\rtextSessionID\x18\x03 \x03(\x03\"A\n\rMatchResponse\x12\x30\n\x0fsentenceMatches\x18\x01 \x03(\x0b\x32\x17.lt_ml_server.MatchList\"1\n\tMatchList\x12$\n\x07matches\x18\x01 \x03(\x0b\x32\x13.lt_ml_server.Match\"\xae\x03\n\x05Match\x12\x0e\n\x06offset\x18\x01 \x01(\r\x12\x0e\n\x06length\x18\x02 \x01(\r\x12\n\n\x02id\x18\x03 \x01(\t\x12\x0e\n\x06sub_id\x18\x04 \x01(\t\x12\x13\n\x0bsuggestions\x18\x05 \x03(\t\x12\x17\n\x0fruleDescription\x18\x06 \x01(\t\x12\x18\n\x10matchDescription\x18\x07 \x01(\t\x12\x1d\n\x15matchShortDescription\x18\x08 \x01(\t\x12\x0b\n\x03url\x18\t \x01(\t\x12\x41\n\x15suggestedReplacements\x18\n \x03(\x0b\x32\".lt_ml_server.SuggestedReplacement\x12\x13\n\x0b\x61utoCorrect\x18\x0b \x01(\x08\x12+\n\x04type\x18\x0c \x01(\x0e\x32\x1d.lt_ml_server.Match.MatchType\x12\x1b\n\x13\x63ontextForSureMatch\x18\r \x01(\x11\x12 \n\x04rule\x18\x0e \x01(\x0b\x32\x12.lt_ml_server.Rule\"1\n\tMatchType\x12\x0f\n\x0bUnknownWord\x10\x00\x12\x08\n\x04Hint\x10\x01\x12\t\n\x05Other\x10\x02\"\xa9\x02\n\x04Rule\x12\x12\n\nsourceFile\x18\x01 \x01(\t\x12\x11\n\tissueType\x18\x02 \x01(\t\x12\x0f\n\x07tempOff\x18\x03 \x01(\x08\x12,\n\x08\x63\x61tegory\x18\x04 \x01(\x0b\x32\x1a.lt_ml_server.RuleCategory\x12\x11\n\tisPremium\x18\x05 \x01(\x08\x12$\n\x04tags\x18\x06 \x03(\x0e\x32\x16.lt_ml_server.Rule.Tag\"\x81\x01\n\x03Tag\x12\t\n\x05picky\x10\x00\x12\x0c\n\x08\x61\x63\x61\x64\x65mic\x10\x01\x12\x0b\n\x07\x63larity\x10\x02\x12\x10\n\x0cprofessional\x10\x03\x12\x0c\n\x08\x63reative\x10\x04\x12\x0c\n\x08\x63ustomer\x10\x05\x12\n\n\x06jobapp\x10\x06\x12\r\n\tobjective\x10\x07\x12\x0b\n\x07\x65legant\x10\x08\"(\n\x0cRuleCategory\x12\n\n\x02id\x18\x01 \x01(\t\x12\x0c\n\x04name\x18\x02 \x01(\t\"\xe2\x01\n\x14SuggestedReplacement\x12\x13\n\x0breplacement\x18\x01 \x01(\t\x12\x13\n\x0b\x64\x65scription\x18\x02 \x01(\t\x12\x0e\n\x06suffix\x18\x03 \x01(\t\x12\x12\n\nconfidence\x18\x04 \x01(\x02\x12?\n\x04type\x18\x05 \x01(\x0e\x32\x31.lt_ml_server.SuggestedReplacement.SuggestionType\";\n\x0eSuggestionType\x12\x0b\n\x07\x44\x65\x66\x61ult\x10\x00\x12\x0f\n\x0bTranslation\x10\x01\x12\x0b\n\x07\x43urated\x10\x02\x32\xa6\x01\n\x10ProcessingServer\x12H\n\x07\x41nalyze\x12\x1c.lt_ml_server.AnalyzeRequest\x1a\x1d.lt_ml_server.AnalyzeResponse\"\x00\x12H\n\x07Process\x12\x1c.lt_ml_server.ProcessRequest\x1a\x1d.lt_ml_server.ProcessResponse\"\x00\x32\xa2\x01\n\x08MLServer\x12\x42\n\x05Match\x12\x1a.lt_ml_server.MatchRequest\x1a\x1b.lt_ml_server.MatchResponse\"\x00\x12R\n\rMatchAnalyzed\x12\".lt_ml_server.AnalyzedMatchRequest\x1a\x1b.lt_ml_server.MatchResponse\"\x00\x32\x65\n\x14PostProcessingServer\x12M\n\x07Process\x12#.lt_ml_server.PostProcessingRequest\x1a\x1b.lt_ml_server.MatchResponse\"\x00\x42*\n\x19org.languagetool.rules.mlB\rMLServerProtob\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'ml_server_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n\031org.languagetool.rules.mlB\rMLServerProto'
  _globals['_ANALYZEREQUEST']._serialized_start=33
  _globals['_ANALYZEREQUEST']._serialized_end=113
  _globals['_PROCESSINGOPTIONS']._serialized_start=116
  _globals['_PROCESSINGOPTIONS']._serialized_end=459
  _globals['_PROCESSINGOPTIONS_LEVEL']._serialized_start=310
  _globals['_PROCESSINGOPTIONS_LEVEL']._serialized_end=459
  _globals['_ANALYZERESPONSE']._serialized_start=461
  _globals['_ANALYZERESPONSE']._serialized_end=529
  _globals['_PROCESSREQUEST']._serialized_start=531
  _globals['_PROCESSREQUEST']._serialized_end=648
  _globals['_PROCESSRESPONSE']._serialized_start=650
  _globals['_PROCESSRESPONSE']._serialized_end=746
  _globals['_ANALYZEDMATCHREQUEST']._serialized_start=748
  _globals['_ANALYZEDMATCHREQUEST']._serialized_end=866
  _globals['_ANALYZEDSENTENCE']._serialized_start=868
  _globals['_ANALYZEDSENTENCE']._serialized_end=953
  _globals['_ANALYZEDTOKENREADINGS']._serialized_start=955
  _globals['_ANALYZEDTOKENREADINGS']._serialized_end=1062
  _globals['_ANALYZEDTOKEN']._serialized_start=1064
  _globals['_ANALYZEDTOKEN']._serialized_end=1125
  _globals['_POSTPROCESSINGREQUEST']._serialized_start=1128
  _globals['_POSTPROCESSINGREQUEST']._serialized_end=1257
  _globals['_MATCHREQUEST']._serialized_start=1259
  _globals['_MATCHREQUEST']._serialized_end=1337
  _globals['_MATCHRESPONSE']._serialized_start=1339
  _globals['_MATCHRESPONSE']._serialized_end=1404
  _globals['_MATCHLIST']._serialized_start=1406
  _globals['_MATCHLIST']._serialized_end=1455
  _globals['_MATCH']._serialized_start=1458
  _globals['_MATCH']._serialized_end=1888
  _globals['_MATCH_MATCHTYPE']._serialized_start=1839
  _globals['_MATCH_MATCHTYPE']._serialized_end=1888
  _globals['_RULE']._serialized_start=1891
  _globals['_RULE']._serialized_end=2188
  _globals['_RULE_TAG']._serialized_start=2059
  _globals['_RULE_TAG']._serialized_end=2188
  _globals['_RULECATEGORY']._serialized_start=2190
  _globals['_RULECATEGORY']._serialized_end=2230
  _globals['_SUGGESTEDREPLACEMENT']._serialized_start=2233
  _globals['_SUGGESTEDREPLACEMENT']._serialized_end=2459
  _globals['_SUGGESTEDREPLACEMENT_SUGGESTIONTYPE']._serialized_start=2400
  _globals['_SUGGESTEDREPLACEMENT_SUGGESTIONTYPE']._serialized_end=2459
  _globals['_PROCESSINGSERVER']._serialized_start=2462
  _globals['_PROCESSINGSERVER']._serialized_end=2628
  _globals['_MLSERVER']._serialized_start=2631
  _globals['_MLSERVER']._serialized_end=2793
  _globals['_POSTPROCESSINGSERVER']._serialized_start=2795
  _globals['_POSTPROCESSINGSERVER']._serialized_end=2896
# @@protoc_insertion_point(module_scope)



from typing import Optional, List
from pydantic import BaseModel, Field, ConfigDict
from enum import Enum

class SuggestionType(int, Enum):
    Default = 0
    Translation = 1
    Curated = 2

class MatchType(int, Enum):
    UnknownWord = 0
    Hint = 1
    Other = 2

class Tag(int, Enum):
    picky = 0
    academic = 1
    clarity = 2
    professional = 3
    creative = 4
    customer = 5
    jobapp = 6
    objective = 7
    elegant = 8

class Software(BaseModel):
    name: Optional[str] = None
    version: Optional[str] = None
    
    model_config = ConfigDict(extra="allow", from_attributes=True)


class IgnoreRange(BaseModel):
    offset: int = Field(ge=0)
    length: int = Field(ge=0)


class SuggestedReplacement(BaseModel):
    replacement: str = ""
    description: Optional[str] = None
    suffix: Optional[str] = None  # Value shown in the UI after the replacement (but not part of it).
    confidence: Optional[float] = None  # from 0 (lowest) to 1 (highest)
    type: SuggestionType = SuggestionType.Default
    model_config = ConfigDict(from_attributes=True)


class Context(BaseModel):
    text: str
    offset: int = Field(ge=0)
    length: int = Field(ge=0)


class RuleCategory(BaseModel):
    id: Optional[str] = None
    name: Optional[str] = None


class RuleTag(BaseModel):
    name: str
    value: str


class RuleUrl(BaseModel):
    value: str = Field(pattern=r"^https?://")  # URI format


class Rule(BaseModel):
    id: Optional[str] = None
    description: Optional[str] = None
    subId: Optional[str] = None
    issueType: Optional[str] = None
    urls: Optional[List[RuleUrl]] = None
    category: Optional[RuleCategory] = None
    tags: Optional[List[Tag]] = None
    sourceFile: Optional[str] = None
    tempOff: Optional[bool] = None
    isPremium: Optional[bool] = None

    model_config = ConfigDict(extra="allow", from_attributes=True)


class Match(BaseModel):
    offset: int = Field(ge=0)
    length: int = Field(ge=0)
    id: Optional[str] = None
    sub_id: Optional[str] = None
    suggestions: Optional[List[str]] = None
    ruleDescription: Optional[str] = None
    matchDescription: Optional[str] = None
    matchShortDescription: Optional[str] = None
    url: Optional[str] = None
    suggestedReplacements: Optional[List[SuggestedReplacement]] = None
    autoCorrect: Optional[bool] = None
    type: Optional[MatchType] = None
    contextForSureMatch: Optional[int] = None
    rule: Optional[Rule] = None
    # For backward compatibility with previous fields
    message: Optional[str] = None
    shortMessage: Optional[str] = None
    suggested_replacements: Optional[List[SuggestedReplacement]] = None
    
    model_config = ConfigDict(extra="allow", from_attributes=True)


class LanguageToolRemoteResult(BaseModel):
    language: str
    languageCode: str
    matches: List[Match]
    languageDetectedCode: Optional[str] = None
    languageDetectedName: Optional[str] = None
    software: Optional[Software] = None
    ignoreRanges: Optional[List[IgnoreRange]] = None
    
    model_config = ConfigDict(extra="allow", from_attributes=True)


import os
from grammared_language.language_tool.output_models import LanguageToolRemoteResult
from grammared_language.utils.errant_grammar_correction_extractor import ErrantGrammarCorrectionExtractor
from functools import lru_cache

class BaseClient:
    def __init__(self, *args, **kwargs):
        self.rule_id = kwargs.get("rule_id", "GrammaredLanguage")
        print(f"Initialized BaseClient with rule_id: {self.rule_id}")
        self.correction_extractor = ErrantGrammarCorrectionExtractor(rule_id=self.rule_id)
    
    def _preprocess(self, text: str) -> str:
        return text
    
    def _predict(self, text: str) -> str:
        raise NotImplementedError("Subclasses should implement this method.")
    
    def _pred_postprocess(self, original: str, pred: str, **kwargs) -> LanguageToolRemoteResult:
        matches = self.correction_extractor.extract_replacements(
            original=original, corrected=pred, fix_tokenization=kwargs.get('fix_tokenization', True)
        )
        return LanguageToolRemoteResult(
            language="English",
            languageCode="en-US",
            matches=matches
        )
    
    def _output_postprocess(self, original: str, pred: LanguageToolRemoteResult, **kwargs) -> LanguageToolRemoteResult:
        return pred

    @lru_cache(maxsize=int(os.getenv('GRAMMARED_LANGUAGE__GRPC_SERVER_CACHE_SIZE', '100000')))
    def predict(self, text: str|list[str]) -> LanguageToolRemoteResult|list[LanguageToolRemoteResult]:
        single = True
        if isinstance(text, list):
            single = False

        if single:
            _text = self._preprocess(text)
        else:
            _text = [self._preprocess(t) for t in text]
        corrected_text: str|list[str] = self._predict(_text)

        output = None
        if single:
            pred: LanguageToolRemoteResult = self._pred_postprocess(text, corrected_text)
            output = self._output_postprocess(text, pred)
        else:
            pred = [
                self._pred_postprocess(orig, corr)
                for orig, corr in zip(text, corrected_text)
            ]
            output = [
                self._output_postprocess(orig, p)
                for orig, p in zip(text, pred)
            ]
        return output

    def __call__(self, text: str) -> LanguageToolRemoteResult:
        return self.predict(text)


# Copyright 2023, NVIDIA CORPORATION & AFFILIATES. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#  * Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#  * Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#  * Neither the name of NVIDIA CORPORATION nor the names of its
#    contributors may be used to endorse or promote products derived
#    from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
# EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
# OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
import os
import json

import numpy as np
import torch
import transformers
import triton_python_backend_utils as pb_utils
from .triton_transformers_model import TritonTransformersPythonModel



class TritonText2TextPythonModel(TritonTransformersPythonModel):
    PIPELINE_TASK = "text2text-generation"



from .base_client import BaseClient
from typing import List, Optional, Union
import json
import numpy as np

try:
    import tritonclient.http as httpclient
    import tritonclient.grpc as grpcclient
    _TRITON_HTTP_AVAILABLE = True
    _TRITON_GRPC_AVAILABLE = True
except Exception:
    try:
        import tritonclient.grpc as grpcclient
        httpclient = None
        _TRITON_HTTP_AVAILABLE = False
        _TRITON_GRPC_AVAILABLE = True
    except Exception:
        httpclient = None
        grpcclient = None
        _TRITON_HTTP_AVAILABLE = False
        _TRITON_GRPC_AVAILABLE = False

_TRITON_AVAILABLE = _TRITON_HTTP_AVAILABLE or _TRITON_GRPC_AVAILABLE

try:
    from transformers import AutoTokenizer
    from grammared_language.grammared_classifier.classifier_pipeline import (
        CalibratedTextClassificationPipeline,
    )
    _TRANSFORMERS_AVAILABLE = True
except Exception:
    AutoTokenizer = None
    CalibratedTextClassificationPipeline = None
    _TRANSFORMERS_AVAILABLE = False

from grammared_language.language_tool.output_models import Match


class GrammarClassificationClient(BaseClient):
    """
    Client for grammar classification.

    Supports two backends:
    - HuggingFace: local `CalibratedTextClassificationPipeline`
    - Triton: remote HTTP inference against a Triton model that returns JSON

    Note: This client overrides `predict` from `BaseClient` to return
    classification outputs (dict or list of dicts) rather than a
    `LanguageToolRemoteResult`.
    """

    def __init__(
        self,
        model_id: str,
        *,
        backend: str = "triton",
        # Triton parameters
        triton_model_name: Optional[str] = None,
        triton_host: str = "localhost",
        triton_port: int = 8001,  # Default to gRPC port
        triton_model_version: Optional[str] = "1",
        triton_protocol: str = "grpc",  # "grpc" or "http"
    ) -> None:
        super().__init__()

        backend = backend.lower()
        if backend not in ("hf", "triton"):
            raise ValueError("backend must be 'hf' or 'triton'")
        self.backend = backend
        self.triton_protocol = triton_protocol.lower()

        if not _TRANSFORMERS_AVAILABLE:
            raise ImportError(
                "transformers is required for GrammarClassificationClient. Install with: pip install transformers"
            )

        self.tokenizer = AutoTokenizer.from_pretrained(model_id)
        self.tokenizer.start_of_replace_token = "<|start_of_replace|>"
        self.tokenizer.end_of_replace_token = "<|end_of_replace|>"

        self.pipeline = None
        self._triton_client = None
        self._triton_model_name = triton_model_name or "grammared_classifier"
        self._triton_model_version = triton_model_version
        self._triton_url = f"{triton_host}:{triton_port}"

        if self.backend == "hf":
            # Initialize tokenizer and calibrated pipeline
            self.pipeline = CalibratedTextClassificationPipeline.from_pretrained(model_id)
        else:
            if not _TRITON_AVAILABLE:
                raise ImportError(
                    "tritonclient is required for Triton backend. Install with: pip install tritonclient[all]"
                )
            
            # Initialize Triton client based on protocol
            if self.triton_protocol == "grpc":
                if not _TRITON_GRPC_AVAILABLE:
                    raise ImportError(
                        "tritonclient.grpc is required for gRPC protocol. Install with: pip install tritonclient[grpc]"
                    )
                self._triton_client = grpcclient.InferenceServerClient(url=self._triton_url)
            else:
                if not _TRITON_HTTP_AVAILABLE:
                    raise ImportError(
                        "tritonclient.http is required for HTTP protocol. Install with: pip install tritonclient[http]"
                    )
                self._triton_client = httpclient.InferenceServerClient(url=self._triton_url)

    def featurizer(self, sentence: str, matches: List[Match], correction_idx: int) -> str:
        """
        Build a single feature string for a given correction index.

        Uses special tokens from the tokenizer:
        - start_of_replace_token, sep_token, end_of_replace_token
        - Prepend cls_token for the selected correction index
        """
        tokenizer = self.tokenizer
        result_segments: List[str] = []
        pointer = 0
        for i, m in enumerate(matches):
            if m.offset > pointer:
                result_segments.append(sentence[pointer:m.offset])

            original = sentence[m.offset : m.offset + m.length]
            replacement = getattr(m, "replacement", None)
            if replacement is None and m.suggestedReplacements and len(m.suggestedReplacements) > 0:
                # Fallback to first suggested replacement text if available
                replacement = m.suggestedReplacements[0].replacement
            if replacement is None:
                replacement = original

            correction_segment = (
                f"{getattr(tokenizer, 'start_of_replace_token', '<|start_of_replace|>')}"
                f"{original}"
                f"{getattr(tokenizer, 'sep_token', '<SEP>')}"
                f"{replacement}"
                f"{getattr(tokenizer, 'end_of_replace_token', '<|end_of_replace|>')}"
            )

            if i == correction_idx:
                correction_segment = f"{getattr(tokenizer, 'cls_token', '<CLS>')}{correction_segment}"

            result_segments.append(correction_segment)
            pointer = m.offset + m.length

        if pointer < len(sentence):
            result_segments.append(sentence[pointer:])
        return "".join(result_segments)

    def predict_matches(self, sentence: str, matches: List[Match]) -> List[dict]:
        """
        Predict classification scores for each correction candidate.

        Returns a list of dicts with keys 'label' and 'score'.
        """
        features = [self.featurizer(sentence, matches, i) for i in range(len(matches))]
        return self.predict(features)  # type: ignore[arg-type]

    def predict(self, texts: Union[str, List[str]]) -> Union[dict, List[dict]]:
        """
        Predict label/score for input text(s).

        - HF backend: uses local calibrated pipeline
        - Triton backend: calls remote Triton HTTP endpoint
        """
        if isinstance(texts, str):
            single = True
            payload = [texts]
        else:
            single = False
            payload = texts

        if self.backend == "hf":
            outputs = self.pipeline.predict(payload, top_k=1)
            return outputs[0] if single else outputs

        # Triton backend
        # Reshape to [batch_size, 1] as expected by the Triton model config
        text_np = np.array(payload, dtype=object).reshape(-1, 1)
        
        # Create inputs based on protocol
        if self.triton_protocol == "grpc":
            inputs = [grpcclient.InferInput("TEXT", list(text_np.shape), "BYTES")]
            inputs[0].set_data_from_numpy(text_np)
            outputs = [grpcclient.InferRequestedOutput("OUTPUT")]
        else:
            inputs = [httpclient.InferInput("TEXT", list(text_np.shape), "BYTES")]
            inputs[0].set_data_from_numpy(text_np)
            outputs = [httpclient.InferRequestedOutput("OUTPUT")]
        
        response = self._triton_client.infer(
            model_name=self._triton_model_name,
            model_version=self._triton_model_version,
            inputs=inputs,
            outputs=outputs,
        )
        out = response.as_numpy("OUTPUT")
        results: List[dict] = []
        for item in out:
            if isinstance(item, bytes):
                s = item.decode("utf-8")
            elif isinstance(item, np.ndarray) and item.dtype.type is np.bytes_:
                s = item.tobytes().decode("utf-8")
            else:
                s = str(item)
            try:
                results.append(json.loads(s))
            except Exception:
                results.append({"label": "unknown", "score": 0.0})

        return results[0] if single else results


"""
Grammared Classifier Model for Triton Inference Server.

This model uses the HuggingFace transformers library with a calibrated
text classification pipeline to classify grammar correction candidates.
"""
import logging
import json
import numpy as np
import triton_python_backend_utils as pb_utils
import torch
from grammared_language.grammared_classifier.classifier_pipeline import CalibratedTextClassificationPipeline
from grammared_language.utils.config_parser import get_model_config

logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO)


class TritonGrammaredClassifierPythonModel:
    """Python model for Grammared Classifier."""

    def initialize(self, args):
        """
        Initialize the model.

        Args:
            args: Dictionary containing model configuration
        """
        logger.warning(f"initialize config: {args}")

        self.model_config = json.loads(args['model_config'])
        logger.info(f"Loaded model config: {self.model_config}")

        self.grammared_language_model_config = json.loads(
            self.model_config.get('parameters', {}).get('grammared_language_model_config', {}).get('string_value', "{}")
        )
        self.grammared_language_model_config = get_model_config(self.model_config.get("name", "grammared_classifier_model"), self.grammared_language_model_config)

        logger.warning(f"Loading grammared_language_model_config: {self.grammared_language_model_config}...")
        # Get model instance device configuration
        model_device_type = args['model_instance_kind']
        model_instance_device_id = args['model_instance_device_id']

        # Determine device
        if self.grammared_language_model_config.serving_config.device == 'cpu':
            self.device = 'cpu'
        elif self.grammared_language_model_config.serving_config.device == 'cuda':
            self.device = f'cuda:{model_instance_device_id}'
        else:
            self.device = 'auto'

        # Initialize attributes for cleanup
        self.pipeline = None
        self.model = None
        self.tokenizer = None

        # Load the HuggingFace model
        if 'pretrained_model_name_or_path' in self.model_config['parameters']:
            model_name = self.model_config['parameters']['pretrained_model_name_or_path']['string_value']
        else:
            raise RuntimeError(f"Failed to get model name from configuration: {self.model_config}")

        try:
            from transformers import AutoTokenizer, AutoModelForSequenceClassification

            logger.info(f"Loading model: {model_name}")
            use_fast=True
            if "deberta-v" in model_name.lower():
                use_fast=False  # Some models have issues with fast tokenizer
            self.tokenizer = AutoTokenizer.from_pretrained(model_name, use_fast=use_fast)
            self.model = AutoModelForSequenceClassification.from_pretrained(model_name, device_map=self.device,)
            self.device = self.model.device

            # # Move model to device
            # self.model.to(self.device)
            self.model.eval()

            # Create calibrated pipeline
            self.pipeline = CalibratedTextClassificationPipeline(
                model=self.model,
                tokenizer=self.tokenizer,
                # device=self.device
            )

            logger.info(f"Successfully loaded model {model_name} on device {self.device}")

            # Log calibration info
            if self.pipeline.calibrators:
                logger.info(f"Model has {len(self.pipeline.calibrators)} calibrators")
            else:
                logger.info("Model has no calibrators - using uncalibrated outputs")

        except Exception as e:
            raise RuntimeError(f"Failed to load model {model_name}: {str(e)}")

    def execute(self, requests):
        """
        Execute inference on a batch of requests.

        Args:
            requests: List of pb_utils.InferenceRequest

        Returns:
            List of pb_utils.InferenceResponse
        """
        responses = []

        for request in requests:
            try:
                # Get input text tensor
                text_tensor = pb_utils.get_input_tensor_by_name(request, "TEXT")

                # Convert to numpy array and decode bytes to strings
                text_np = text_tensor.as_numpy()

                # Handle both single string and batch of strings
                # Triton sends strings as byte arrays
                texts = []
                for text_bytes in text_np:
                    if isinstance(text_bytes, bytes):
                        text_str = text_bytes.decode('utf-8')
                    else:
                        text_str = str(text_bytes)
                    texts.append(text_str)

                logger.debug(f"Processing {len(texts)} text(s)")

                # Tokenize the texts
                encoded = self.tokenizer(
                    texts,
                    padding=True,
                    truncation=True,
                    max_length=512,
                    return_tensors="pt"
                )

                # Move to device
                input_ids = encoded['input_ids'].to(self.device)
                attention_mask = encoded['attention_mask'].to(self.device)

                # Run inference through the model
                with torch.no_grad():
                    outputs = self.model(
                        input_ids=input_ids,
                        attention_mask=attention_mask
                    )

                # Get logits
                logits = outputs.logits
                logits_np = logits.cpu().numpy()

                # Apply calibration
                calibrated_probs = self.pipeline._apply_calibration(logits_np)

                # Get label mappings
                if hasattr(self.model.config, 'id2label') and self.model.config.id2label:
                    id2label = self.model.config.id2label
                else:
                    id2label = {i: f"LABEL_{i}" for i in range(calibrated_probs.shape[1])}

                # Format outputs as label + score dictionaries
                results = []
                for probs in calibrated_probs:
                    # Get the predicted class (highest probability)
                    predicted_idx = int(np.argmax(probs))
                    predicted_label = id2label[predicted_idx]
                    predicted_score = float(probs[predicted_idx])

                    result = {
                        'label': predicted_label,
                        'score': predicted_score
                    }
                    results.append(result)

                # Convert results to JSON strings
                json_results = [json.dumps(result) for result in results]

                # Create numpy array of JSON strings
                # For single input, return one result; for batch, return list
                output_json_np = np.array(json_results, dtype=object)

                # Create output tensor
                output_json = pb_utils.Tensor(
                    "OUTPUT",
                    output_json_np
                )

                # Create inference response
                response = pb_utils.InferenceResponse(
                    output_tensors=[output_json]
                )

            except Exception as e:
                # Return error response
                error_msg = f"Error during inference: {str(e)}"
                logger.error(error_msg, exc_info=True)
                response = pb_utils.InferenceResponse(
                    output_tensors=[],
                    error=pb_utils.TritonError(error_msg)
                )

            responses.append(response)

        return responses

    def finalize(self):
        """Clean up resources."""
        # Clean up model safely
        if hasattr(self, 'pipeline') and self.pipeline is not None:
            del self.pipeline
        if hasattr(self, 'model') and self.model is not None:
            del self.model
        if hasattr(self, 'tokenizer') and self.tokenizer is not None:
            del self.tokenizer
        if torch.cuda.is_available():
            torch.cuda.empty_cache()
        logger.info("Model finalized and resources cleaned up")



from .base_client import BaseClient
from typing import Optional
import numpy as np
import jinja2
import logging

# Configure logging
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)

try:
    import tritonclient.http as httpclient
    import tritonclient.grpc as grpcclient
    _TRITON_HTTP_AVAILABLE = True
    _TRITON_GRPC_AVAILABLE = True
except Exception:
    try:
        import tritonclient.grpc as grpcclient
        httpclient = None
        _TRITON_HTTP_AVAILABLE = False
        _TRITON_GRPC_AVAILABLE = True
    except Exception:
        httpclient = None
        grpcclient = None
        _TRITON_HTTP_AVAILABLE = False
        _TRITON_GRPC_AVAILABLE = False

_TRITON_AVAILABLE = _TRITON_HTTP_AVAILABLE or _TRITON_GRPC_AVAILABLE


class Text2TextBaseClient(BaseClient):
    """
    Client for text-to-text generation models on Triton Inference Server.
    
    This client provides a common interface for LLM/text2text models that:
    - Accept a text input string
    - Generate corrected/transformed text output
    - Can be used for grammar correction, text editing, etc.
    
    Example models: CoEdit, T5, FLAN-T5, etc.
    
    Args:
        model_name: Name of the Triton model to use
        triton_host: Triton server host (default: "localhost")
        triton_port: Triton server port (default: 8001 for gRPC, 8000 for HTTP)
        triton_model_version: Model version (default: "1")
        triton_protocol: Communication protocol - "grpc" or "http" (default: "grpc")
        input_name: Name of the input tensor (default: "text_input")
        output_name: Name of the output tensor (default: "text_output")
        prompt_template: Optional template string for formatting input (Jinja2-compatible).
                      Use {{ text }} as placeholder. Example: "Fix grammar: {{ text }}"
                      For chat-style models, use format like: "<|user|>\n{{ text }}<|assistant|>\n"
    """
    
    def __init__(
        self,
        model_name: str,
        *,
        triton_host: str = "localhost",
        triton_port: int = 8001,  # Default to gRPC port
        triton_model_version: str = "1",
        triton_protocol: str = "grpc",  # "grpc" or "http"
        input_name: str = "text_input",
        output_name: str = "text_output",
        prompt_template: Optional[str] = None,
        **kwargs
    ):
        if "rule_id" not in kwargs:
            kwargs["rule_id"] = model_name
        super().__init__(**kwargs)
        
        if not _TRITON_AVAILABLE:
            raise ImportError(
                "tritonclient is required for Text2TextBaseClient. "
                "Install with: pip install tritonclient[all]"
            )
        
        self.model_name = model_name
        self.triton_model_version = triton_model_version
        self.triton_protocol = triton_protocol.lower()
        self.input_name = input_name
        self.output_name = output_name
        self.prompt_template = prompt_template
        self._template = jinja2.Template(prompt_template) if prompt_template else None
        
        
        # Initialize Triton client based on protocol
        triton_url = f"{triton_host}:{triton_port}"
        logger.warning(f"triton_url: {triton_url}")
        if self.triton_protocol == "grpc":
            if not _TRITON_GRPC_AVAILABLE:
                raise ImportError(
                    "tritonclient.grpc is required for gRPC protocol. "
                    "Install with: pip install tritonclient[grpc]"
                )
            self._triton_client = grpcclient.InferenceServerClient(url=triton_url)
        else:
            if not _TRITON_HTTP_AVAILABLE:
                raise ImportError(
                    "tritonclient.http is required for HTTP protocol. "
                    "Install with: pip install tritonclient[http]"
                )
            self._triton_client = httpclient.InferenceServerClient(url=triton_url)
        
    def _preprocess(self, text: str) -> str:
        """Apply prompt template if configured."""
        if self._template:
            return self._template.render(text=text)
        return text
    
    def _predict(self, text: str|list[str]) -> str|list[str]:
        """
        Send text to Triton model and get generated output.
        
        Args:
            text: Input text (already preprocessed)
            
        Returns:
            Generated/corrected text from the model
        """
        # Prepare input as numpy array with shape [1, 1] for batching
        # First dimension is batch size, second is the string dimension
        single = True
        if isinstance(text, list):
            text_np = np.array([text], dtype=object)
            single = False
        else:
            text_np = np.array([[text]], dtype=object)
        
        # Create Triton input/output tensors based on protocol
        if self.triton_protocol == "grpc":
            inputs = [
                grpcclient.InferInput(
                    self.input_name, 
                    list(text_np.shape), 
                    "BYTES"
                )
            ]
            inputs[0].set_data_from_numpy(text_np)
            outputs = [grpcclient.InferRequestedOutput(self.output_name)]
        else:
            inputs = [
                httpclient.InferInput(
                    self.input_name, 
                    list(text_np.shape), 
                    "BYTES"
                )
            ]
            inputs[0].set_data_from_numpy(text_np)
            outputs = [httpclient.InferRequestedOutput(self.output_name)]
        
        # Send inference request
        response = self._triton_client.infer(
            model_name=self.model_name,
            model_version=self.triton_model_version,
            inputs=inputs,
            outputs=outputs,
        )
        
        # Extract output text
        output_data = response.as_numpy(self.output_name)
        
        # Handle different output formats
        if len(output_data) == 0:
            return text  # Return original if no output
        
        # Extract from batch: output_data has shape [batch_size, 1]
        def decode_result(result):
            if isinstance(result, bytes):
                return result.decode("utf-8")
            elif isinstance(result, np.ndarray):
                if result.dtype.type is np.bytes_:
                    return result.tobytes().decode("utf-8")
                return str(result)
            else:
                return str(result)
        results = [decode_result(output_data[i][0]) for i in range(len(output_data))]
        if single:
            return results[0]
        return results


from typing import Dict, Any, Optional, Union, Literal
from pydantic import BaseModel, Field

class ModelConfig(BaseModel):
    """Configuration for an individual model.

    Flexible configuration that supports various model types and backends.
    All fields are optional or have sensible defaults to maximize flexibility.
    """
    # Core fields - flexible typing
    type: Optional[str] = Literal["gector", "coedit", "deliterater", "grammared_classifier", "llm"]
    backend: Optional[str] = Literal["triton", "local"]

    # Optional fields
    pretrained_model_name_or_path: Optional[str] = Field(
        None, description="HuggingFace model name or local path to pretrained model"
    )
    triton_model_name: Optional[str] = Field(
        None, description="Name of the model in Triton server"
    )
    triton_hostname: Optional[str] = Field(
        None, description="Hostname of the Triton server"
    )
    triton_port: Optional[int] = Field(None, description="Port of the Triton server")

    # Flexible configuration dictionaries
    model_kwargs: Optional[Dict[str, Any]] = Field(
        None, description="Additional model-specific arguments (e.g., temperature, max_new_tokens)"
    )
    configs: Optional[Any] = Field(
        None, description="Model-specific configurations"
    )
    rule_generation: Optional[Any] = Field(
        None, description="Rule generation configuration (if applicable)"
    )

    class Config:
        extra = "allow"  # Allow any additional fields


class ModelsConfig(BaseModel):
    """Top-level configuration containing all models.

    Flexible structure that allows for various model configurations.
    """
    models: Dict[str, Union[ModelConfig, Dict[str, Any]]] = Field(
        default_factory=dict, description="Dictionary of model configurations keyed by model name"
    )

    class Config:
        extra = "allow"  # Allow additional top-level fields

# Convenience alias for the main config
Config = ModelsConfig




"""Grammared Language - Grammar correction tools and models."""

import importlib
import pkgutil
from pathlib import Path

__version__ = "0.1.0"

# Programmatically discover and import all submodules
__all__ = ["__version__"]
_module_path = Path(__file__).parent

# Import subpackages
for _, module_name, is_pkg in pkgutil.iter_modules([str(_module_path)]):
    if not module_name.startswith('_') and is_pkg:
        try:
            module = importlib.import_module(f'.{module_name}', package=__name__)
            globals()[module_name] = module
            __all__.append(module_name)
        except ImportError as e:
            # Skip modules that have missing dependencies
            pass



import numpy as np
from transformers import TextClassificationPipeline
from typing import Dict, List, Union, Optional
import torch


def load_logistic_weights(weight_dict, lr_model=None):
    """Load weights & bias into a sklearn LogisticRegression from a dict."""
    if lr_model is None:
        from sklearn.linear_model import LogisticRegression
        lr_model = LogisticRegression()
    lr_model.coef_ = np.array(weight_dict["coef"])
    lr_model.intercept_ = np.array(weight_dict["intercept"])
    lr_model.classes_ = np.array(weight_dict["classes"])
    return lr_model


class CalibratedTextClassificationPipeline(TextClassificationPipeline):
    """
    Text classification pipeline with probability calibration support.

    Loads calibrator weights from model. config. probability_calibrator_weights
    and applies calibration to the model outputs.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.calibrators = None
        self._load_calibrators()

    def _load_calibrators(self):
        """Load calibrators from model config if available."""
        if hasattr(self, 'model') and hasattr(self.model, 'config') and hasattr(self.model.config, 'probability_calibrator_weights'):
            calibrator_weights = self.model.config.probability_calibrator_weights
            if calibrator_weights:
                self.calibrators = [
                    load_logistic_weights(weights)
                    for weights in calibrator_weights
                ]
                print(f"Loaded {len(self.calibrators)} calibrators from model config")
        else:
            print("No calibrators found in model config - using uncalibrated outputs")

    def _apply_calibration(self, logits: np.ndarray) -> np.ndarray:
        """
        Apply calibration to logits.

        Args:
            logits: Raw model logits, shape (batch_size, num_classes)

        Returns:
            Calibrated probabilities, shape (batch_size, num_classes)
        """
        if self.calibrators is None:
            # Fallback to softmax if no calibrators
            from scipy.special import softmax
            return softmax(logits, axis=1)

        # Apply each calibrator to get calibrated probability for each class
        calibrated_probs = np.column_stack([
            calibrator.predict_proba(logits)[:, 1]  # Probability for positive class
            for calibrator in self.calibrators
        ])

        # Normalize to sum to 1
        calibrated_probs = calibrated_probs / calibrated_probs.sum(axis=1, keepdims=True)

        return calibrated_probs

    def postprocess(self, model_outputs, function_to_apply=None, top_k=1, **kwargs):
        """
        Postprocess model outputs with calibration.

        Args:
            model_outputs: Raw model outputs containing logits
            function_to_apply: Optional function to apply (overridden by calibration)
            top_k: Number of top predictions to return

        Returns:
            List of dictionaries containing labels and calibrated scores
        """
        # Extract logits
        if isinstance(model_outputs, dict):
            logits = model_outputs["logits"]
        else:
            logits = model_outputs[0]

        # Convert to numpy
        if isinstance(logits, torch.Tensor):
            logits = logits.detach().cpu().numpy()

        # Apply calibration
        calibrated_probs = self._apply_calibration(logits)

        # Get label mappings
        if self.model.config.id2label:
            id2label = self.model.config.id2label
        else:
            id2label = {i: f"LABEL_{i}" for i in range(calibrated_probs.shape[1])}

        # Format outputs
        outputs = []
        for probs in calibrated_probs:
            # Get top-k predictions
            top_indices = np.argsort(probs)[::-1][:top_k]

            result = [
                {
                    "label": id2label[int(idx)],
                    "score": float(probs[idx])
                }
                for idx in top_indices
            ]

            outputs.append(result if top_k > 1 else result[0])

        return outputs[0]

    def predict_proba(self, texts: Union[str, List[str]]) -> np.ndarray:
        """
        Get calibrated probability predictions for texts.

        Args:
            texts: Single text or list of texts

        Returns:
            Calibrated probabilities, shape (num_texts, num_classes)
        """
        # Ensure texts is a list
        if isinstance(texts, str):
            texts = [texts]
            single_input = True
        else:
            single_input = False

        # Get model outputs
        outputs = self(texts, top_k=self.model.config.num_labels)

        # Extract probabilities
        if single_input:
            outputs = [outputs]

        probs = []
        for output in outputs:
            if single_input:
                output = output[0]
            sorted_output = sorted(output,
                                   key=lambda x: self.model.config.label2id[x['label']])
            probs.append([item['score'] for item in sorted_output])

        probs = np.array(probs)

        return probs[0] if single_input else probs

    def predict(self, texts: Union[str, List[str]], return_dict=True, **kwargs) -> Union[str, List[dict], dict]:
        """
        Get calibrated predicted labels for texts.

        Args:
            texts: Single text or list of texts
            return_dict: return dict with score or not
            **kwargs: Additional arguments passed to the pipeline call (e.g., top_k)

        Returns:
            Predicted label(s)
        """
        # Ensure texts is a list
        if isinstance(texts, str):
            texts = [texts]
            single_input = True
        else:
            single_input = False

        # Get predictions
        top_k = kwargs.get('top_k', 1)
        outputs = self(texts, top_k=top_k)

        if not return_dict:
            outputs = [o['label'] for o in outputs]

        # Extract labels
        if single_input:
            return outputs[0]
        else:
            return outputs

    def get_uncalibrated_proba(self, texts: Union[str, List[str]]) -> np.ndarray:
        """
        Get uncalibrated probability predictions (raw softmax).

        Args:
            texts: Single text or list of texts

        Returns:
            Uncalibrated probabilities, shape (num_texts, num_classes)
        """
        # Temporarily disable calibrators
        calibrators_backup = self.calibrators
        self.calibrators = None

        try:
            probs = self.predict_proba(texts)
        finally:
            # Restore calibrators
            self.calibrators = calibrators_backup

        return probs

    @classmethod
    # Convenience function to create the pipeline
    def from_pretrained(
            cls,
            model_name_or_path: str,
            **kwargs
    ):
        """
        Create a calibrated text classification pipeline.

        Args:
            model_name_or_path: HuggingFace model identifier or local path
            **kwargs:  Additional arguments passed to pipeline

        Returns:
            CalibratedTextClassificationPipeline instance

        Example:
            >>> pipeline = CalibratedTextClassificationPipeline.from_pretrained("rayliuca/grammared-classifier-deberta-v3-base")
            >>> predictions = pipeline.predict(["This is great!", "This is bad!"])
            >>> probs = pipeline.predict_proba(["This is great! "])
        """
        from transformers import AutoTokenizer, AutoModelForSequenceClassification

        tokenizer = AutoTokenizer.from_pretrained(model_name_or_path)
        model = AutoModelForSequenceClassification.from_pretrained(model_name_or_path)

        return cls(
            model=model,
            tokenizer=tokenizer,
            **kwargs
        )


"""
Grammar Error Correction - Extract Replacements
Extracts replacements between original and corrected text for grammar correction.
"""
from typing import List, Tuple
import re
from difflib import SequenceMatcher

from grammared_language.language_tool.output_models import Match, SuggestedReplacement, MatchType, SuggestionType

class GrammarCorrectionExtractor:
    """Extract replacement operations from original to corrected text."""

    def __init__(self, min_length: int = 1):
        """
        Initialize the extractor.

        Args:
            min_length: Minimum length of original text to be replaced (must be > 0)
        """
        self.min_length = max(1, min_length)  # Ensure at least 1

    def extract_replacements(self, original: str, corrected: str, fix_tokenization=True) -> list[Match]:
        """
        Extract replacements needed to transform original text to corrected text.

        Args:
            original: The original text with errors
            corrected: The corrected text

        Returns:
            List of replacements, each containing:
            - offset: int (start index in original text)
            - length: int (length of text to replace in original)
            - replacement: str (corrected text to insert)
        """
        # Pre-process to fix common tokenization mistakes in corrected text
        if fix_tokenization:
            fixed_corrected = self._fix_tokenization_mistakes(corrected)
        else:
            fixed_corrected = corrected

        # Tokenize into words while preserving positions
        orig_tokens, orig_positions = self._tokenize_with_positions(original)
        corr_tokens, corr_positions = self._tokenize_with_positions(fixed_corrected)

        matches = []

        # Use SequenceMatcher on tokens
        matcher = SequenceMatcher(None, orig_tokens, corr_tokens)
        opcodes = list(matcher.get_opcodes())

        # Track which opcodes have been processed to avoid double-processing
        consumed_indices = set()

        for idx, (tag, i1, i2, j1, j2) in enumerate(opcodes):
            # Skip if this opcode was already consumed by a previous operation
            if idx in consumed_indices:
                continue

            if tag == 'replace':
                # Get character positions
                offset = orig_positions[i1]
                end_pos = orig_positions[i2 - 1] + len(orig_tokens[i2 - 1])
                length = end_pos - offset
                mistake = {"offset": offset, "length": length, "text": original[offset:end_pos]}
                # Get replacement text from corrected
                if j2 > j1:
                    repl_start = corr_positions[j1]
                    repl_end = corr_positions[j2 - 1] + len(corr_tokens[j2 - 1])
                    replacement = fixed_corrected[repl_start:repl_end]
                else:
                    replacement = ""

                # Apply constraints:
                # - length must be >= min_length
                # - replacement cannot be empty
                if length >= self.min_length and replacement != "":
                    matches.append(
                        Match(
                            message=replacement,
                            suggested_replacements=[
                                SuggestedReplacement(
                                    replacement=replacement,
                                    # type=SuggestionType.Translation
                                )
                            ],
                            offset=offset,
                            length=length,

                            # /** Spelling errors, typically red. */
                            # UnknownWord = 0;
                            # /** Style errors, typically light blue. */
                            # Hint = 1;
                            # /** Other errors (including grammar), typically yellow/orange. */
                            # Other = 2;
                            # !!! not used/ accepted by LanguageTool Server
                            type=MatchType.Hint,
                            # rule=Rule(
                            #     id=rule_id,
                            #     description=rule_description
                            # )
                        )
                    )

            elif tag == 'insert':
                # Convert insertion to replacement by including adjacent token
                # or by merging with a following delete operation

                # Get the inserted text
                repl_start = corr_positions[j1]
                repl_end = corr_positions[j2 - 1] + len(corr_tokens[j2 - 1])
                inserted_text = fixed_corrected[repl_start:repl_end]

                # Try to merge with consecutive operations (delete/replace)
                converted = self._try_merge_consecutive_operations(
                    original, orig_tokens, orig_positions, corr_tokens, corr_positions,
                    fixed_corrected, i1, None, inserted_text, opcodes, idx
                )

                if converted:
                    match, consumed_idx = converted
                    matches.append(match)
                    if consumed_idx is not None:
                        consumed_indices.add(consumed_idx)

            elif tag == 'delete':
                # Try to merge delete with following insert, or with adjacent token
                converted = self._try_merge_consecutive_operations(
                    original, orig_tokens, orig_positions, corr_tokens, corr_positions,
                    fixed_corrected, i1, i2, None, opcodes, idx
                )

                if converted:
                    match, consumed_idx = converted
                    matches.append(match)
                    if consumed_idx is not None:
                        consumed_indices.add(consumed_idx)

        return matches

    def _try_merge_consecutive_operations(self, original: str, orig_tokens: List[str],
                                          orig_positions: List[int], corr_tokens: List[str],
                                          corr_positions: List[int], fixed_corrected: str,
                                          start_idx: int, end_idx: int, inserted_text: str,
                                          opcodes: List, current_idx: int) -> tuple:
        """
        Attempt to merge consecutive operations or with adjacent tokens.
        Handles:
        - Insert + next delete/replace -> include next token
        - Delete + next insert/replace -> merge into replacement
        - Standalone insert/delete -> merge with adjacent token

        Args:
            original: Original text
            orig_tokens: Original tokens
            orig_positions: Original token positions
            corr_tokens: Corrected tokens
            corr_positions: Corrected token positions
            fixed_corrected: Fixed corrected text
            start_idx: Start token index in original
            end_idx: End token index in original (None for insert)
            inserted_text: Text being inserted (None for delete)
            opcodes: All opcodes from SequenceMatcher
            current_idx: Current opcode index

        Returns:
            Tuple of (Match object, consumed_opcode_index) or None if no merge possible
        """
        is_insert = end_idx is None
        is_delete = inserted_text is None

        # Check if there's a next operation we can merge with
        if current_idx + 1 < len(opcodes):
            next_tag, next_i1, next_i2, next_j1, next_j2 = opcodes[current_idx + 1]

            # Case 1: insert followed by delete/replace at the same position
            if is_insert and next_tag in ('delete', 'replace') and next_i1 == start_idx:
                next_token = orig_tokens[start_idx]
                replacement = inserted_text + " " + next_token
                match = self._create_match(
                    orig_positions[start_idx], len(next_token), replacement
                )[0]
                return (match, current_idx + 1)

            # Case 2: delete followed by insert/replace at the same position
            if is_delete and next_tag in ('insert', 'replace') and next_i1 == end_idx:
                offset = orig_positions[start_idx]
                end_pos = orig_positions[end_idx - 1] + len(orig_tokens[end_idx - 1])
                length = end_pos - offset

                # Get the replacement text from corrected
                repl_start = corr_positions[next_j1]
                repl_end = corr_positions[next_j2 - 1] + len(corr_tokens[next_j2 - 1])
                replacement = fixed_corrected[repl_start:repl_end]

                if length >= self.min_length and replacement != "":
                    match = self._create_match(offset, length, replacement)[0]
                    return (match, current_idx + 1)

        # No consecutive merge possible, try merging with adjacent tokens
        if is_insert:
            return self._merge_with_adjacent_token(
                original, orig_tokens, orig_positions,
                start_idx, inserted_text, opcodes, current_idx, is_insert=True
            )
        else:
            return self._merge_with_adjacent_token(
                original, orig_tokens, orig_positions,
                start_idx, end_idx, opcodes, current_idx, is_insert=False
            )

    def _create_match(self, offset: int, length: int, replacement: str) -> tuple:
        """Helper to create a Match object with the given parameters."""
        return (Match(
            message=replacement,
            suggested_replacements=[
                SuggestedReplacement(replacement=replacement)
            ],
            offset=offset,
            length=length,
        ), None)

    def _merge_with_adjacent_token(self, original: str, orig_tokens: List[str],
                                   orig_positions: List[int], start_idx: int,
                                   end_idx_or_text: int, opcodes: List, current_idx: int,
                                   is_insert: bool) -> tuple:
        """
        Merge an insert or delete with an adjacent token to create a valid replacement.

        Args:
            original: Original text
            orig_tokens: Original tokens
            orig_positions: Original token positions
            start_idx: Start token index
            end_idx_or_text: End index (for delete) or inserted text (for insert)
            opcodes: All opcodes
            current_idx: Current opcode index
            is_insert: True if this is an insert, False if delete

        Returns:
            Tuple of (Match, consumed_idx) or None
        """
        if is_insert:
            inserted_text = end_idx_or_text
            # Try next token first
            if start_idx < len(orig_tokens):
                next_token_used = False
                if current_idx + 1 < len(opcodes):
                    next_tag, next_i1, _, _, _ = opcodes[current_idx + 1]
                    if next_tag in ('replace', 'delete') and next_i1 == start_idx:
                        next_token_used = True

                if not next_token_used:
                    next_token = orig_tokens[start_idx]
                    replacement = inserted_text + " " + next_token
                    return self._create_match(
                        orig_positions[start_idx], len(next_token), replacement
                    )

            # Try previous token
            if start_idx > 0:
                prev_token = orig_tokens[start_idx - 1]
                replacement = prev_token + " " + inserted_text
                return self._create_match(
                    orig_positions[start_idx - 1], len(prev_token), replacement
                )
        else:
            # Delete case
            del_end = end_idx_or_text
            # Try previous token
            if start_idx > 0:
                prev_idx = start_idx - 1
                prev_token = orig_tokens[prev_idx]
                end_pos = orig_positions[del_end - 1] + len(orig_tokens[del_end - 1])
                length = end_pos - orig_positions[prev_idx]

                if length >= self.min_length:
                    return self._create_match(
                        orig_positions[prev_idx], length, prev_token
                    )

            # Try next token
            if del_end < len(orig_tokens):
                offset = orig_positions[start_idx]
                next_token = orig_tokens[del_end]
                end_pos = orig_positions[del_end] + len(next_token)
                length = end_pos - offset

                if length >= self.min_length:
                    return self._create_match(offset, length, next_token)

        return None

    def _fix_tokenization_mistakes(self, text: str) -> str:
        """
        Fix common tokenization mistakes from generation models.

        Common mistakes:
        - "I ' m" -> "I'm" (spaces around apostrophe in contractions)
        - "word ." -> "word." (space before period)
        - "word ," -> "word," (space before comma)
        - "word ?" -> "word?" (space before question mark)
        - "word !" -> "word!" (space before exclamation)
        - "word ;" -> "word;" (space before semicolon)
        - "word :" -> "word:" (space before colon)

        Args:
            text: Input text with potential tokenization mistakes

        Returns:
            Fixed text with proper spacing
        """
        result = text

        # Fix spaces around apostrophes in contractions
        # Pattern 1: letter + space(s) + ' + space(s) + letter (e.g., "I ' m")
        result = re.sub(r"(\w)\s+'\s+(\w)", r"\1'\2", result)
        # Pattern 2: letter + ' + space(s) + letter (e.g., "I 'm")
        result = re.sub(r"(\w)'\s+(\w)", r"\1'\2", result)
        # Pattern 3: letter + space(s) + ' + letter (e.g., "it 's" - less common)
        result = re.sub(r"(\w)\s+'(\w)", r"\1'\2", result)

        # Fix space before punctuation
        # Period - don't add extra space
        result = re.sub(r'\s+\.', '.', result)
        # Comma
        result = re.sub(r'\s+,', ',', result)
        # Question mark
        result = re.sub(r'\s+\?', '?', result)
        # Exclamation mark
        result = re.sub(r'\s+!', '!', result)
        # Semicolon
        result = re.sub(r'\s+;', ';', result)
        # Colon (but be careful not to break things like "http : //")
        result = re.sub(r'(\w)\s+:', r'\1:', result)

        # Fix space after opening quotes and before closing quotes
        # But preserve the space between the quote and the surrounding words
        result = re.sub(r'"\s+(\S)', r'" \1', result)
        result = re.sub(r'(\S)\s+"', r'\1 "', result)

        # Fix spaces around hyphens in compound words (letter-letter only)
        # This preserves numeric ranges (10 - 20) and math operations (5 - 3)
        result = re.sub(r'([a-zA-Z])\s+-\s+([a-zA-Z])', r'\1-\2', result)

        return result

    def _tokenize_with_positions(self, text: str) -> Tuple[List[str], List[int]]:
        """
        Tokenize text into words while preserving character positions.

        Args:
            text: Input text

        Returns:
            Tuple of (tokens, positions) where positions[i] is the character offset of tokens[i]
        """
        tokens = []
        positions = []
        current_token = ""
        token_start = 0

        for i, char in enumerate(text):
            if char.isspace():
                if current_token:
                    tokens.append(current_token)
                    positions.append(token_start)
                    current_token = ""
            else:
                if not current_token:
                    token_start = i
                current_token += char

        # Add last token
        if current_token:
            tokens.append(current_token)
            positions.append(token_start)

        return tokens, positions


from pathlib import Path
from urllib.request import urlopen

from .base_client import BaseClient
from grammared_language.language_tool.output_models import LanguageToolRemoteResult
from grammared_language.utils.errant_grammar_correction_extractor import ErrantGrammarCorrectionExtractor

try:
    from gector import GECToR, predict, load_verb_dict
    GECTOR_AVAILABLE = True
except ImportError:
    GECToR = None
    predict = None
    load_verb_dict = None
    GECTOR_AVAILABLE = False

try:
    from gector import GECToRTriton
    GECTOR_TRITON_AVAILABLE = True
except (ImportError, AttributeError):
    GECToRTriton = None
    GECTOR_TRITON_AVAILABLE = False

try:
    from transformers import AutoTokenizer
    TRANSFORMERS_AVAILABLE = True
except ImportError:
    AutoTokenizer = None
    TRANSFORMERS_AVAILABLE = False

OFFICIAL_VOCAB_URL = "https://github.com/grammarly/gector/raw/master/data/verb-form-vocab.txt"
class GectorClient(BaseClient):
    def __init__(
            self,
            pretrained_model_name_or_path: str,
            triton_model_name: str=None,
            triton_host: str='localhost',
            triton_port: int=8001,
            verb_dict_path: str='data/verb-form-vocab.txt',
            auto_download_official_vocab: bool=True,
            **kwargs):
        if "rule_id" not in kwargs:
            kwargs["rule_id"] = triton_model_name or pretrained_model_name_or_path or "Gector"
        super().__init__(**kwargs)
        
        if not GECTOR_AVAILABLE:
            raise ImportError("gector package is required for GectorClient. Install it with: pip install gector")
        
        if not TRANSFORMERS_AVAILABLE:
            raise ImportError("transformers package is required for GectorClient. Install it with: pip install transformers")
        
        if triton_model_name is None:
            self.model = GECToR.from_pretrained(pretrained_model_name_or_path)
        else:
            if GECToRTriton is None:
                raise ImportError("GECToRTriton is not available in your gector installation. Please upgrade gector or use the standard model.")
            self.model = GECToRTriton.from_pretrained(
                pretrained_model_name_or_path, 
                model_name=triton_model_name,
                triton_url=f"{triton_host}:{triton_port}"
            )

        self.tokenizer = AutoTokenizer.from_pretrained(pretrained_model_name_or_path)
        verb_dict_file = Path(verb_dict_path)
        if auto_download_official_vocab and not verb_dict_file.is_file():
            verb_dict_file.parent.mkdir(parents=True, exist_ok=True)
            try:
                with urlopen(OFFICIAL_VOCAB_URL) as response:
                    verb_dict_file.write_bytes(response.read())
            except Exception as exc:
                raise FileNotFoundError(
                    f"Failed to download verb vocab from {OFFICIAL_VOCAB_URL} to {verb_dict_file}"
                ) from exc

        self.encode, self.decode = load_verb_dict(verb_dict_path)

        self.pred_config = {
            'keep_confidence': kwargs.get('keep_confidence', 0),
            'min_error_prob': kwargs.get('min_error_prob', 0),
            'n_iteration': kwargs.get('n_iteration', 5),
            'batch_size': kwargs.get('batch_size', 2),
        }

    def _predict(self, text: str|list[str], **kwargs) -> str|list[str]:
        if isinstance(text, list):
            corrected = predict(
                self.model, self.tokenizer, text,
                self.encode, self.decode,
                keep_confidence=kwargs.get('keep_confidence', self.pred_config['keep_confidence']),
                min_error_prob=kwargs.get('min_error_prob', self.pred_config['min_error_prob']),
                n_iteration=kwargs.get('n_iteration', self.pred_config['n_iteration']),
                batch_size=kwargs.get('batch_size', self.pred_config['batch_size']),
            )
            return corrected
        
        corrected = predict(
            self.model, self.tokenizer, [text],
            self.encode, self.decode,
            keep_confidence=kwargs.get('keep_confidence', self.pred_config['keep_confidence']),
            min_error_prob=kwargs.get('min_error_prob', self.pred_config['min_error_prob']),
            n_iteration=kwargs.get('n_iteration', self.pred_config['n_iteration']),
            batch_size=kwargs.get('batch_size', self.pred_config['batch_size']),
        )
        return corrected[0]


"""Utilities for parsing model configuration files."""

from pathlib import Path
from typing import Dict, Any, Optional, List, Literal, Union
import yaml
import os
from pydantic import BaseModel, ConfigDict, Field, field_validator

from grammared_language.clients.base_client import BaseClient

DEFAULT_MODEL_CONFIG_PATH = "/default_model_config.yaml"
MODEL_CONFIG_PATH = "/model_config.yaml"
MODEL_REPO_FOLDER = "/models"

import logging
# Configure logging
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


class ServingConfig(BaseModel):
    """Configuration for model serving settings."""
    
    model_config = ConfigDict(extra='allow')
    
    triton_host: Optional[str] = 'localhost'
    triton_port: Optional[int] = 8001
    triton_model_name: Optional[str] = None
    triton_protocol: Literal['grpc', 'http'] = 'grpc'  # 'grpc' or 'http'
    pretrained_model_name_or_path: Optional[str] = None
    backend: Optional[str] = "transformers" # "transformers", "ort"
    device: Optional[str] = "auto"  # "cpu", "cuda", or "auto"
    

class ModelInitConfig(BaseModel):
    """Configuration for model inference settings."""
    
    model_config = ConfigDict(extra='allow')
    


class ModelInferenceConfig(BaseModel):
    """Configuration for model inference settings."""
    
    model_config = ConfigDict(extra='allow')
    
    # temperature: Optional[float] = None
    # max_length: Optional[int] = None
    # num_beams: Optional[int] = None


class GrammaredConfig(BaseModel):
    """Configuration for grammared-specific settings."""
    
    model_config = ConfigDict(extra='allow')
    
    prompt_template: Optional[str] = None
    error_classifier: Optional[str] = None


class BaseModelConfig(BaseModel):
    """Base configuration for all model types."""
    
    model_config = ConfigDict(extra='allow')  # Allow additional fields for model-specific configs
    
    type: str
    backend: Literal[
        'triton', 
        # 'local', # Local inference
        'openai' # OpenAI compatiable api
        ] = 'triton'
    
    # Nested config format
    serving_config: ServingConfig
    model_init_config: Optional[ModelInitConfig] = Field(default=ModelInitConfig(), alias='model_config')
    model_inference_config: Optional[ModelInferenceConfig] = ModelInferenceConfig()
    grammared_config: Optional[GrammaredConfig] = GrammaredConfig()

class GectorConfig(BaseModelConfig):
    """Configuration for GECToR models."""
    
    type: Literal['gector'] = 'gector'


class GrammaredClassifierConfig(BaseModelConfig):
    """Configuration for Grammared Classifier models."""
    
    type: Literal['grammared_classifier'] = 'grammared_classifier'


class Text2TextInferenceConfig(BaseModel):
    """Configuration for Text2Text model inference settings."""
    
    backend: Literal['transformers', 'ort'] = 'transformers'
    
    model_config = ConfigDict(extra='allow')


class Text2TextConfig(BaseModelConfig):
    """Configuration for Text2Text models."""
    
    type: Literal['text2text'] = 'text2text'
    config: Optional[Text2TextInferenceConfig] = None

class CoEditConfig(Text2TextConfig):
    """Configuration for CoEdit models."""
    
    type: Literal['coedit'] = 'coedit'


class OpenAIConfig(BaseModelConfig):
    """Configuration for OpenAI compatible models."""
    
    type: Literal['openai'] = 'openai'

    openai_base_url: Optional[str] = "http://localhost:11434/v1/"
    openai_api_key: Optional[str] = "ollama"


MODEL_CONFIG_REGISTRY = {
    'gector': GectorConfig,
    'grammared_classifier': GrammaredClassifierConfig,
    'coedit': CoEditConfig,
    'text2text': Text2TextConfig,
    'openai': OpenAIConfig,
}

def get_model_config(model_name, model_config_dict: Dict[str, Any]) -> BaseModelConfig:
    model_type = model_config_dict.get('type')
    if model_type not in MODEL_CONFIG_REGISTRY:
        raise ValueError(f"Unknown model type for {model_name}: {model_type}")
    return MODEL_CONFIG_REGISTRY[model_type](**model_config_dict)

class ModelsConfig(BaseModel):
    """Container for multiple model configurations."""
    
    models: Dict[str, Union[GectorConfig, GrammaredClassifierConfig, CoEditConfig, BaseModelConfig]]
    
    @classmethod
    def from_dict(cls, config_dict: Dict[str, Any]) -> 'ModelsConfig':
        """Create ModelsConfig from a dictionary."""
        models = {}
        for model_name, model_config in config_dict.items():
            models[model_name] = get_model_config(model_name, model_config)
        return cls(models=models)


def load_config_from_file(config_path: str) -> ModelsConfig:
    """
    Load configuration from a YAML file.
    
    Args:
        config_path: Path to the configuration file
        
    Returns:
        ModelsConfig containing validated model configurations
        
    Raises:
        FileNotFoundError: If the config file doesn't exist
    """
    config_file = Path(config_path)
    if not config_file.exists():
        raise FileNotFoundError(f"Config file not found: {config_path}")
    
    with open(config_file, 'r') as f:
        config_dict = yaml.safe_load(f)
    config = ModelsConfig.from_dict(config_dict)
    logger.warning(config)
    return config


def load_config_from_env(prefix: str = "GRAMMARED_LANGUAGE") -> ModelsConfig:
    """
    Load configuration from environment variables.
    
    Environment variables should follow the pattern:
    {PREFIX}__MODELS__{MODEL_NAME}__{KEY}={VALUE}
    
    Example:
        GRAMMARED_LANGUAGE__MODELS__GECTOR_DEBERTA_LARGE__TYPE=gector
        GRAMMARED_LANGUAGE__MODELS__GECTOR_DEBERTA_LARGE__BACKEND=triton
        GRAMMARED_LANGUAGE__MODELS__GECTOR_DEBERTA_LARGE__SERVING_CONFIG__TRITON_HOSTNAME=localhost
    
    Args:
        prefix: Environment variable prefix (default: "GRAMMARED_LANGUAGE")
        
    Returns:
        ModelsConfig containing validated model configurations
        
    Raises:
        ValueError: If no matching environment variables are found
    """
    prefix_with_sep = f"{prefix}__"
    config_dict: Dict[str, Any] = {}
    
    # Find all environment variables with the prefix
    env_vars = {k: v for k, v in os.environ.items() if k.startswith(prefix_with_sep)}
    
    if not env_vars:
        raise ValueError(f"No environment variables found with prefix: {prefix_with_sep}")
    
    for env_key, env_value in env_vars.items():
        # Remove prefix and split by double underscore
        key_parts = env_key[len(prefix_with_sep):].split("__")
        
        # Convert keys to lowercase for model names
        key_parts = [part.lower() for part in key_parts]
        
        # Build nested dictionary
        current = config_dict
        for i, part in enumerate(key_parts[:-1]):
            if part not in current:
                current[part] = {}
            current = current[part]
        
        # Set the final value, attempting type conversion
        final_key = key_parts[-1]
        
        # Try to convert value to appropriate type
        try:
            # Try boolean
            if env_value.lower() in ('true', 'false'):
                current[final_key] = env_value.lower() == 'true'
            # Try integer
            elif env_value.isdigit():
                current[final_key] = int(env_value)
            # Try float
            elif '.' in env_value and env_value.replace('.', '').replace('-', '').isdigit():
                current[final_key] = float(env_value)
            # Keep as string
            else:
                current[final_key] = env_value
        except (ValueError, AttributeError):
            current[final_key] = env_value
    
    # Extract models dict (skip the "models" key level)
    models_dict = config_dict.get("models", {})
    
    if not models_dict:
        raise ValueError(f"No models configuration found in environment variables")
    
    logger.info(f"Loaded configuration for {len(models_dict)} model(s) from environment variables")
    
    return ModelsConfig.from_dict(models_dict)


def create_client_from_config(
    model_name: str, 
    config: Union[GectorConfig, GrammaredClassifierConfig, CoEditConfig, Dict[str, Any]]
) -> Optional[BaseClient]:
    """
    Create a client instance from configuration.
    
    Args:
        model_name: Name of the model
        config: Configuration (Pydantic model or dictionary)
        
    Returns:
        Initialized client instance or None if creation fails
    """
    # Convert dict to Pydantic model if needed
    if isinstance(config, dict):
        model_type = config.get('type')
        if model_type not in MODEL_CONFIG_REGISTRY:
            raise ValueError(f"Unknown model type for {model_name}: {model_type}")
        config = MODEL_CONFIG_REGISTRY[model_type](**config)
    
    try:
        # Import appropriate client class based on type
        if isinstance(config, GectorConfig):
            from grammared_language.clients.gector_client import GectorClient
            
            # GectorClient expects model_id and triton_model_name
            client_params = vars(config.serving_config)
            client_params.update(vars(config.grammared_config) if config.grammared_config else {})
            
            return GectorClient(**client_params)
        
        elif isinstance(config, GrammaredClassifierConfig):
            from grammared_language.clients.grammar_classification_client import GrammarClassificationClient
                        
            # Filter out non-client parameters
            client_params = vars(config.serving_config)
            client_params.update(vars(config.grammared_config) if config.grammared_config else {})

            logger.warning(f"Creating GrammarClassificationClient with params: {client_params}")
            return GrammarClassificationClient(**client_params)
            
        elif isinstance(config, CoEditConfig):
            from grammared_language.clients.coedit_client import CoEditClient
            
            client_params = vars(config.serving_config)
            client_params.update(vars(config.grammared_config) if config.grammared_config else {})
            
            logger.warning(f"Creating CoEditClient with params: {client_params}")
            return CoEditClient(**client_params)
            
    except Exception as e:
        logger.error(f"Failed to create client for {model_name}: {e}")
    
    return None


"""
1. check if MODEL_REPO_FOLDER exists
2. check MODEL_CONFIG_PATH
3. check environment variable
4. use DEFAULT_MODEL_CONFIG_PATH
"""
def get_config(config_path:str=MODEL_CONFIG_PATH, use_env: bool=True, backup_config_path=DEFAULT_MODEL_CONFIG_PATH) -> ModelsConfig:
    env_vars = {k: v for k, v in os.environ.items() if k.startswith("GRAMMARED_LANGUAGE__")}
    # Determine which config path to use
    if os.path.isfile(config_path):
        try:
            logger.info(f"Loading model configuration from file: {config_path}")
            config = load_config_from_file(config_path)
            return config
        except Exception as e:
            logger.warning(f"Failed to load config from file {config_path}: {e}")

    if env_vars and use_env:
        try:
            logger.info(f"Loading model configuration from environment variables")
            config = load_config_from_env()
            return config
        except Exception as e:
            logger.warning(f"Failed to load config from environment variables: {e}")

    logger.info(f"No config path provided. Loading model configuration from backup file: {backup_config_path}")
    config = load_config_from_file(backup_config_path)
    return config


def create_clients_from_config(config_path: str=None, use_env: bool=True, backup_config_path: str="/default_model_config.yaml") -> List[BaseClient]:
    """
    Initialize multiple clients from a configuration file or environment variables.
    
    Args:
        config_path: Path to the configuration file
        
    Returns:
        List of initialized client instances
    """
    models_config = get_config(config_path=config_path, use_env=use_env, backup_config_path=backup_config_path)
    clients = []
    
    for m_name, m_config in models_config.models.items():
        client = create_client_from_config(m_name, m_config)
        if client:
            clients.append(client)
    
    return clients



"""
Auto-generated gRPC protocol buffer code from LanguageTool's ml_server.proto.

This module contains auto-generated code from ml_server.proto.
The files are named grammar_check_pb2* because that's what protoc generates
from the protobuf package name, but they implement the ml_server protocol.

Do not manually edit files in this module - regenerate using:

    python generate_grpc.py

"""

from . import ml_server_pb2
from . import ml_server_pb2_grpc

__all__ = ['ml_server_pb2', 'ml_server_pb2_grpc']





# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc
import warnings

from . import ml_server_pb2 as ml__server__pb2

GRPC_GENERATED_VERSION = '1.67.1'
GRPC_VERSION = grpc.__version__
_version_not_supported = False

try:
    from grpc._utilities import first_version_is_lower
    _version_not_supported = first_version_is_lower(GRPC_VERSION, GRPC_GENERATED_VERSION)
except ImportError:
    _version_not_supported = True

if _version_not_supported:
    raise RuntimeError(
        f'The grpc package installed is at version {GRPC_VERSION},'
        + f' but the generated code in ml_server_pb2_grpc.py depends on'
        + f' grpcio>={GRPC_GENERATED_VERSION}.'
        + f' Please upgrade your grpc module to grpcio>={GRPC_GENERATED_VERSION}'
        + f' or downgrade your generated code using grpcio-tools<={GRPC_VERSION}.'
    )


class ProcessingServerStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.Analyze = channel.unary_unary(
                '/lt_ml_server.ProcessingServer/Analyze',
                request_serializer=ml__server__pb2.AnalyzeRequest.SerializeToString,
                response_deserializer=ml__server__pb2.AnalyzeResponse.FromString,
                _registered_method=True)
        self.Process = channel.unary_unary(
                '/lt_ml_server.ProcessingServer/Process',
                request_serializer=ml__server__pb2.ProcessRequest.SerializeToString,
                response_deserializer=ml__server__pb2.ProcessResponse.FromString,
                _registered_method=True)


class ProcessingServerServicer(object):
    """Missing associated documentation comment in .proto file."""

    def Analyze(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Process(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_ProcessingServerServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'Analyze': grpc.unary_unary_rpc_method_handler(
                    servicer.Analyze,
                    request_deserializer=ml__server__pb2.AnalyzeRequest.FromString,
                    response_serializer=ml__server__pb2.AnalyzeResponse.SerializeToString,
            ),
            'Process': grpc.unary_unary_rpc_method_handler(
                    servicer.Process,
                    request_deserializer=ml__server__pb2.ProcessRequest.FromString,
                    response_serializer=ml__server__pb2.ProcessResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'lt_ml_server.ProcessingServer', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))
    server.add_registered_method_handlers('lt_ml_server.ProcessingServer', rpc_method_handlers)


 # This class is part of an EXPERIMENTAL API.
class ProcessingServer(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def Analyze(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/lt_ml_server.ProcessingServer/Analyze',
            ml__server__pb2.AnalyzeRequest.SerializeToString,
            ml__server__pb2.AnalyzeResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def Process(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/lt_ml_server.ProcessingServer/Process',
            ml__server__pb2.ProcessRequest.SerializeToString,
            ml__server__pb2.ProcessResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)


class MLServerStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.Match = channel.unary_unary(
                '/lt_ml_server.MLServer/Match',
                request_serializer=ml__server__pb2.MatchRequest.SerializeToString,
                response_deserializer=ml__server__pb2.MatchResponse.FromString,
                _registered_method=True)
        self.MatchAnalyzed = channel.unary_unary(
                '/lt_ml_server.MLServer/MatchAnalyzed',
                request_serializer=ml__server__pb2.AnalyzedMatchRequest.SerializeToString,
                response_deserializer=ml__server__pb2.MatchResponse.FromString,
                _registered_method=True)


class MLServerServicer(object):
    """Missing associated documentation comment in .proto file."""

    def Match(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def MatchAnalyzed(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_MLServerServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'Match': grpc.unary_unary_rpc_method_handler(
                    servicer.Match,
                    request_deserializer=ml__server__pb2.MatchRequest.FromString,
                    response_serializer=ml__server__pb2.MatchResponse.SerializeToString,
            ),
            'MatchAnalyzed': grpc.unary_unary_rpc_method_handler(
                    servicer.MatchAnalyzed,
                    request_deserializer=ml__server__pb2.AnalyzedMatchRequest.FromString,
                    response_serializer=ml__server__pb2.MatchResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'lt_ml_server.MLServer', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))
    server.add_registered_method_handlers('lt_ml_server.MLServer', rpc_method_handlers)


 # This class is part of an EXPERIMENTAL API.
class MLServer(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def Match(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/lt_ml_server.MLServer/Match',
            ml__server__pb2.MatchRequest.SerializeToString,
            ml__server__pb2.MatchResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def MatchAnalyzed(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/lt_ml_server.MLServer/MatchAnalyzed',
            ml__server__pb2.AnalyzedMatchRequest.SerializeToString,
            ml__server__pb2.MatchResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)


class PostProcessingServerStub(object):
    """for e.g. resorting suggestions
    """

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.Process = channel.unary_unary(
                '/lt_ml_server.PostProcessingServer/Process',
                request_serializer=ml__server__pb2.PostProcessingRequest.SerializeToString,
                response_deserializer=ml__server__pb2.MatchResponse.FromString,
                _registered_method=True)


class PostProcessingServerServicer(object):
    """for e.g. resorting suggestions
    """

    def Process(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_PostProcessingServerServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'Process': grpc.unary_unary_rpc_method_handler(
                    servicer.Process,
                    request_deserializer=ml__server__pb2.PostProcessingRequest.FromString,
                    response_serializer=ml__server__pb2.MatchResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'lt_ml_server.PostProcessingServer', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))
    server.add_registered_method_handlers('lt_ml_server.PostProcessingServer', rpc_method_handlers)


 # This class is part of an EXPERIMENTAL API.
class PostProcessingServer(object):
    """for e.g. resorting suggestions
    """

    @staticmethod
    def Process(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/lt_ml_server.PostProcessingServer/Process',
            ml__server__pb2.PostProcessingRequest.SerializeToString,
            ml__server__pb2.MatchResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)



from ..grammared_classifier.classifier_pipeline import CalibratedTextClassificationPipeline
from ..language_tool.output_models import Match, SuggestedReplacement

class TritonGrammaredClassifierModel:
    def __init__(self, model_name: str, tokenizer=None):
        self.pipeline = CalibratedTextClassificationPipeline.from_pretrained(model_name)

        if tokenizer is None:
            from transformers import AutoTokenizer
            tokenizer = AutoTokenizer.from_pretrained(model_name)
        self.tokenizer = tokenizer

    def featurizer(self, sentence: str, matches: list[Match], correction_idx:int) -> str:
        tokenizer = self.tokenizer
        sentence = sentence
        # Sort corrections in case they're out of order
        # sorted_matches = sorted(matches, key=lambda x: x.offset)
        result_segments = []
        pointer = 0
        for i, m in enumerate(matches):
            # Add unchanged text before correction
            if m.offset > pointer:
                result_segments.append(sentence[pointer:m.offset])

            original = sentence[m.offset:m.offset + m.length]
            replacement = m.replacement
            # Build correction segment
            correction_segment = (
                f"{tokenizer.start_of_replace_token}"
                f"{original}"
                f"{tokenizer.sep_token}"
                f"{replacement}"
                f"{tokenizer.end_of_replace_token}"
            )
            # Prepend cls_token ONLY if this is the selected correction
            if i == correction_idx:
                correction_segment = f"{tokenizer.cls_token}{correction_segment}"
            result_segments.append(correction_segment)
            pointer = m.offset + m.length
        # Add any trailing text after last correction
        if pointer < len(sentence):
            result_segments.append(sentence[pointer:])
        return ''.join(result_segments)

    # def predict(self, sentence, matches: list[Match]) -> list:
    #     features = [self.featurizer(sentence, matches, i) for i in range(len(matches))]
    #     predictions = self.pipeline.predict(features, top_k=1)

    def predict(self, sentence_input: str|list[str]) -> list:
        if isinstance(sentence_input, str):
            return self.pipeline.predict(sentence_input, top_k=1)
        return self.pipeline.predict(sentence_input, top_k=1)


from .text2text_base_client import Text2TextBaseClient
from typing import Optional
import logging

# Configure logging
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


class CoEditClient(Text2TextBaseClient):
    """
    Client for CoEdit grammar correction models on Triton Inference Server.
    
    CoEdit is a text editing model trained for grammar correction and text refinement.
    This client provides a convenient interface with sensible defaults for CoEdit models.
    
    Model: grammarly/coedit-large, grammarly/coedit-xl
    Paper: https://arxiv.org/abs/2305.09857
    
    Args:
        model_name: Name of the Triton model (default: "coedit_large")
        triton_host: Triton server host (default: "localhost")
        triton_port: Triton server port (default: 8001 for gRPC, 8000 for HTTP)
        triton_model_version: Model version (default: "1")
        triton_protocol: Communication protocol - "grpc" or "http" (default: "grpc")
        prompt_template: Optional template string for formatting input (Jinja2-compatible).
                      Use {{ text }} as placeholder. Defaults to "Fix grammatical errors: {{ text }}".
                      Example: "Fix grammar: {{ text }}"
                      For chat-style: "<|user|>\n{{ text }}<|assistant|>\n"
        **kwargs: Additional arguments passed to Text2TextBaseClient
    
    Example:
        >>> client = CoEditClient()
        >>> result = client.predict("She go to the store yesterday.")
        >>> print(result.matches)
    """

    DEFAULT_PROMPT_TEMPLATE = "Fix grammatical errors: {{ text }}"
    
    def __init__(
        self,
        model_name: str = "coedit_large",
        triton_host: str = "localhost",
        triton_port: int = 8001,  # Default to gRPC port
        triton_model_version: str = "1",
        triton_protocol: str = "grpc",  # "grpc" or "http"
        prompt_template: str = DEFAULT_PROMPT_TEMPLATE,
        **kwargs
    ):
        logger.warning({
            "model_name": model_name,
            "triton_host": triton_host,
            "triton_port": triton_port,
            "triton_model_version": triton_model_version,
            "triton_protocol": triton_protocol,
            "prompt_template": prompt_template,
            "kwargs": kwargs,
        })
        # Initialize parent with CoEdit-specific defaults
        super().__init__(
            model_name=model_name,
            triton_host=triton_host,
            triton_port=triton_port,
            triton_model_version=triton_model_version,
            triton_protocol=triton_protocol,
            input_name="text_input",
            output_name="text_output",
            prompt_template=prompt_template,
            **kwargs
        )



"""
Grammar Error Correction - Extract Replacements using ERRANT
Extracts replacements between original and corrected text using the ERRANT library.
"""
from typing import List
import re
import errant

from grammared_language.language_tool.output_models import Match, SuggestedReplacement, MatchType


ERROR_TYPE_LABELS = {
    "NOUN:POSS": "Noun Possessive",
    "CONTR": "Contraction",
    "VERB:FORM": "Verb Form",
    "VERB:TENSE": "Verb Tense",
    "NOUN": "Noun Choice",
    "VERB": "Verb Choice",
    "ADJ": "Adjective",
    "ADV": "Adverb",
    "PREP": "Preposition",
    "DET": "Determiner / Article",
    "PART": "Particle / Phrasal Verb",
    "PUNCT": "Punctuation",
    "SPACE": "Whitespace",
    "PRON": "Pronoun",
    "CONJ": "Conjunction",
    "NUM": "Number",
    "SYM": "Symbol",
    "X": "Other Token",
    "OTHER": "Other / Uncategorized",
    "ORTH": "Orthography (Case / Space)",
    "WO": "Word Order",
    "SPELL": "Spelling",
    "MORPH": "Morphology / Derivation",
    "ADJ:FORM": "Adjective Form",
    "NOUN:NUM": "Noun Number",
    "VERB:SVA": "Subject–Verb Agreement",
}

class ErrantGrammarCorrectionExtractor:
    """Extract replacement operations from original to corrected text using ERRANT."""

    def __init__(self, language: str = 'en', min_length: int = 1, rule_id:str="grammared_language") -> None:
        """
        Initialize the extractor with ERRANT annotator.

        Args:
            language: Language code for ERRANT (default: 'en')
            min_length: Minimum length of original text to be replaced (must be > 0)
        """
        self.annotator = errant.load(language)
        self.min_length = max(1, min_length)
        self.rule_id = rule_id

    def extract_replacements(self, original: str, corrected: str, fix_tokenization=True) -> List[Match]:
        """
        Extract replacements needed to transform original text to corrected text.

        Args:
            original: The original text with errors
            corrected: The corrected text
            fix_tokenization: Whether to fix common tokenization mistakes in corrected text

        Returns:
            List of Match objects containing:
            - offset: int (start character index in original text)
            - length: int (length of text to replace in original)
            - message: str (error type from ERRANT)
            - suggested_replacements: List[SuggestedReplacement] (corrected text)
            - type: MatchType (hint for grammar errors)
        """
        # Pre-process to fix common tokenization mistakes in corrected text
        if fix_tokenization:
            fixed_corrected = self._fix_tokenization_mistakes(corrected)
        else:
            fixed_corrected = corrected
        
        # Parse texts with ERRANT
        orig_parsed = self.annotator.parse(original)
        cor_parsed = self.annotator.parse(fixed_corrected)
        
        # Get edits from ERRANT
        edits = self.annotator.annotate(orig_parsed, cor_parsed)
        
        matches = []
        
        for edit in edits:
            # Convert token indices to character indices
            offset, length = self._token_span_to_char_span(
                orig_parsed, edit.o_start, edit.o_end
            )
            
            # Get the replacement text (use character-based string from edit)
            replacement = edit.c_str
            error_type = edit.type
            
            # Handle insertions (where o_start == o_end, length == 0)
            # Convert to a replacement by including adjacent token
            if length == 0 and replacement.strip():
                # This is an insertion - need to convert to replacement with context
                offset, length, replacement = self._convert_insertion_to_replacement(
                    orig_parsed, original, edit.o_start, replacement
                )
            
            # Apply constraints:
            # - length must be >= min_length
            # - replacement cannot be empty (unless it's a deletion, which we skip)
            if length >= self.min_length and replacement.strip():
                # Create error message from ERRANT error type
                if ":" in error_type:
                    error_type = ":".join(error_type.split(":")[1:])
                else:
                    error_type = "OTHER"
                error_type = ERROR_TYPE_LABELS.get(error_type, error_type)
                original_text = edit.o_str if edit.o_str else f"[insert at position {edit.o_start}]"
                message = error_type
                
                matches.append(
                    Match(
                        message=message,
                        suggested_replacements=[
                            SuggestedReplacement(
                                replacement=replacement.strip(),
                                description=error_type
                            )
                        ],
                        offset=offset,
                        length=length,
                        type=MatchType.Other,  # Grammar errors typically map to "Other". Tho this has no effect in LT.
                        id=self.rule_id
                    )
                )
        
        return matches

    def _token_span_to_char_span(self, parsed_doc, token_start: int, token_end: int) -> tuple[int, int]:
        """
        Convert token span to character span.

        Args:
            parsed_doc: Spacy Doc object from ERRANT
            token_start: Start token index
            token_end: End token index

        Returns:
            Tuple of (char_offset, char_length)
        """
        if token_start == token_end:
            # Empty span - use position at token_start
            if token_start < len(parsed_doc):
                offset = parsed_doc[token_start].idx
                return offset, 0
            else:
                # End of document
                if len(parsed_doc) > 0:
                    last_token = parsed_doc[-1]
                    offset = last_token.idx + len(last_token.text)
                else:
                    offset = 0
                return offset, 0
        
        # Get character offset of first token
        start_token = parsed_doc[token_start]
        offset = start_token.idx
        
        # Get character offset of last token + its length
        end_token = parsed_doc[token_end - 1]
        end_char = end_token.idx + len(end_token.text)
        
        # Calculate length
        length = end_char - offset
        
        return offset, length

    def _convert_insertion_to_replacement(self, parsed_doc, original_text: str, 
                                         insert_position: int, inserted_text: str) -> tuple[int, int, str]:
        """
        Convert an insertion to a replacement by including an adjacent token.
        
        Args:
            parsed_doc: Spacy Doc object from ERRANT
            insert_position: Token position where text should be inserted
            inserted_text: The text being inserted
            original_text: Original text string
            
        Returns:
            Tuple of (offset, length, replacement_with_context)
        """
        # Try to include the following token (preferred)
        if insert_position < len(parsed_doc):
            token = parsed_doc[insert_position]
            offset = token.idx
            length = len(token.text)
            # Build replacement: inserted text + space + following token
            replacement = f"{inserted_text} {token.text}"
            return offset, length, replacement
        
        # If at end, try previous token
        if insert_position > 0:
            token = parsed_doc[insert_position - 1]
            offset = token.idx
            length = len(token.text)
            # Build replacement: previous token + space + inserted text
            replacement = f"{token.text} {inserted_text}"
            return offset, length, replacement
        
        # Edge case: empty document or single insertion
        # Return the insertion at position 0
        return 0, 0, inserted_text

    def _fix_tokenization_mistakes(self, text: str) -> str:
        """
        Fix common tokenization mistakes from generation models.

        Common mistakes:
        - "I ' m" -> "I'm" (spaces around apostrophe in contractions)
        - "word ." -> "word." (space before period)
        - "word ," -> "word," (space before comma)
        - "word ?" -> "word?" (space before question mark)
        - "word !" -> "word!" (space before exclamation)
        - "word ;" -> "word;" (space before semicolon)
        - "word :" -> "word:" (space before colon)

        Args:
            text: Input text with potential tokenization mistakes

        Returns:
            Fixed text with proper spacing
        """
        result = text

        # Fix spaces around apostrophes in contractions
        # Pattern 1: letter + space(s) + ' + space(s) + letter (e.g., "I ' m")
        result = re.sub(r"(\w)\s+'\s+(\w)", r"\1'\2", result)
        # Pattern 2: letter + ' + space(s) + letter (e.g., "I 'm")
        result = re.sub(r"(\w)'\s+(\w)", r"\1'\2", result)
        # Pattern 3: letter + space(s) + ' + letter (e.g., "it 's" - less common)
        result = re.sub(r"(\w)\s+'(\w)", r"\1'\2", result)

        # Fix space before punctuation
        # Period - don't add extra space
        result = re.sub(r'\s+\.', '.', result)
        # Comma
        result = re.sub(r'\s+,', ',', result)
        # Question mark
        result = re.sub(r'\s+\?', '?', result)
        # Exclamation mark
        result = re.sub(r'\s+!', '!', result)
        # Semicolon
        result = re.sub(r'\s+;', ';', result)
        # Colon (but be careful not to break things like "http : //")
        result = re.sub(r'(\w)\s+:', r'\1:', result)

        # Fix space after opening quotes and before closing quotes
        # But preserve the space between the quote and the surrounding words
        result = re.sub(r'"\s+(\S)', r'" \1', result)
        result = re.sub(r'(\S)\s+"', r'\1 "', result)

        # Fix spaces around hyphens in compound words (letter-letter only)
        # This preserves numeric ranges (10 - 20) and math operations (5 - 3)
        result = re.sub(r'([a-zA-Z])\s+-\s+([a-zA-Z])', r'\1-\2', result)

        return result
